import numpy as np
import pytest
from astropy import units as u
from astropy.table import Table
from astropy.tests.helper import assert_quantity_allclose
from astropy.utils.exceptions import AstropyUserWarning
from numpy.testing import assert_allclose
from regions import RectanglePixelRegion, PixCoord

from jdaviz.core.custom_units_and_equivs import PIX2, SPEC_PHOTON_FLUX_DENSITY_UNITS
from jdaviz.core.unit_conversion_utils import (flux_conversion_general,
                                               handle_squared_flux_unit_conversions)


def test_cubeviz_aperphot_cube_orig_flux(cubeviz_helper, image_cube_hdu_obj_microns):
    cubeviz_helper.load_data(image_cube_hdu_obj_microns, data_label="test")
    flux_unit = u.Unit("1E-17 erg*s^-1*cm^-2*Angstrom^-1*pix^-2")
    solid_angle_unit = PIX2

    aper = RectanglePixelRegion(center=PixCoord(x=1, y=2), width=3, height=5)

    cubeviz_helper.plugins['Subset Tools'].import_region(aper)

    # Make sure MASK is not an option even when shown in viewer.
    cubeviz_helper.app.add_data_to_viewer("flux-viewer", "test[MASK]", visible=True)

    plg = cubeviz_helper.plugins["Aperture Photometry"]._obj
    assert plg.dataset.labels == ["test[FLUX]", "test[ERR]"]
    assert plg.cube_slice == "4.894e+00 um"

    plg.dataset_selected = "test[FLUX]"
    plg.aperture_selected = "Subset 1"
    plg.vue_do_aper_phot()
    row = plg.export_table()[0]

    # Basically, we should recover the input rectangle here.
    assert_allclose(row["xcenter"], 1 * u.pix)
    assert_allclose(row["ycenter"], 2 * u.pix)
    sky = row["sky_center"]
    assert_allclose(sky.ra.deg, 205.43985906934287)
    assert_allclose(sky.dec.deg, 27.003490103642033)

    # sum should be in flux ( have factor of pix^2 multiplied out of input unit)
    assert_allclose(row["sum"], 75 * flux_unit * solid_angle_unit)  # 3 (w) x 5 (h) x 5 (v)

    assert_allclose(row["sum_aper_area"], 15 * solid_angle_unit)  # 3 (w) x 5 (h)
    assert_allclose(row["mean"], 5 * flux_unit)
    assert_quantity_allclose(row["slice_wave"], 4.894499866699333 * u.um)

    # Move slider and make sure it recomputes for a new slice automatically.
    cube_slice_plg = cubeviz_helper.plugins["Slice"]._obj
    cube_slice_plg.vue_goto_first()
    plg.vue_do_aper_phot()
    row = plg.export_table()[1]

    # Same rectangle but different slice value.
    assert_allclose(row["xcenter"], 1 * u.pix)
    assert_allclose(row["ycenter"], 2 * u.pix)
    sky = row["sky_center"]
    assert_allclose(sky.ra.deg, 205.43985906934287)
    assert_allclose(sky.dec.deg, 27.003490103642033)

    # sum should be in flux ( have factor of pix^2 multiplied out of input unit)
    assert_allclose(row["sum"], 15 * flux_unit * solid_angle_unit)  # 3 (w) x 5 (h) x 1 (v)

    assert_allclose(row["sum_aper_area"], 15 * solid_angle_unit)  # 3 (w) x 5 (h)
    assert_allclose(row["mean"], 1 * flux_unit)
    assert_quantity_allclose(row["slice_wave"], 4.8904998665093435 * u.um)

    # We continue on with test_cubeviz_aperphot_generated_2d_collapse here
    # because we want to make sure the result would append properly between 3D and 2D.
    collapse_plg = cubeviz_helper.plugins["Collapse"]._obj
    collapse_plg.vue_collapse()

    # Need this to make it available for photometry data drop-down.
    cubeviz_helper.app.add_data_to_viewer("uncert-viewer", "test[FLUX] collapsed")

    plg = cubeviz_helper.plugins["Aperture Photometry"]._obj
    plg.dataset_selected = "test[FLUX] collapsed"
    plg.aperture_selected = "Subset 1"
    plg.vue_do_aper_phot()
    row = plg.export_table()[2]

    # Basically, we should recover the input rectangle here.
    assert_allclose(row["xcenter"], 1 * u.pix)
    assert_allclose(row["ycenter"], 2 * u.pix)
    sky = row["sky_center"]
    assert_allclose(sky.ra.deg, 205.43985906934287)
    assert_allclose(sky.dec.deg, 27.003490103642033)

    # sum should be in flux ( have factor of pix^2 multiplied out of input unit)
    assert_allclose(row["sum"], 540 * flux_unit * solid_angle_unit)  # 3 (w) x 5 (h) x 36 (v)

    assert_allclose(row["sum_aper_area"], 15 * solid_angle_unit)  # 3 (w) x 5 (h)
    assert_allclose(row["mean"], 36 * flux_unit)
    assert np.isnan(row["slice_wave"])

    # Invalid counts conversion factor
    plg.counts_factor = -1
    plg.vue_do_aper_phot()
    assert "cannot be negative" in plg.result_failed_msg


def test_cubeviz_aperphot_generated_3d_gaussian_smooth(cubeviz_helper, image_cube_hdu_obj_microns):
    cubeviz_helper.load_data(image_cube_hdu_obj_microns, data_label="test")
    flux_unit = u.Unit("1E-17 erg*s^-1*cm^-2*Angstrom^-1*pix^-2")  # actually a sb
    solid_angle_unit = PIX2

    gauss_plg = cubeviz_helper.plugins["Gaussian Smooth"]._obj
    gauss_plg.mode_selected = "Spatial"
    with pytest.warns(AstropyUserWarning, match="The following attributes were set on the data"):
        _ = gauss_plg.smooth()

    # Need this to make it available for photometry data drop-down.
    cubeviz_helper.app.add_data_to_viewer("uncert-viewer", "test[FLUX] spatial-smooth stddev-1.0")

    aper = RectanglePixelRegion(center=PixCoord(x=1, y=2), width=3, height=5)
    cubeviz_helper.plugins['Subset Tools'].import_region(aper)

    plg = cubeviz_helper.plugins["Aperture Photometry"]._obj
    plg.dataset_selected = "test[FLUX] spatial-smooth stddev-1.0"
    plg.aperture_selected = "Subset 1"
    plg.vue_do_aper_phot()
    row = cubeviz_helper.plugins['Aperture Photometry'].export_table()[0]

    # Basically, we should recover the input rectangle here.
    assert_allclose(row["xcenter"], 1 * u.pix)
    assert_allclose(row["ycenter"], 2 * u.pix)
    sky = row["sky_center"]
    assert_allclose(sky.ra.deg, 205.43985906934287)
    assert_allclose(sky.dec.deg, 27.003490103642033)

    # sum should be in flux ( have factor of pix^2 multiplied out of input unit)
    assert_allclose(row["sum"], 48.54973 * flux_unit * solid_angle_unit)  # 3 (w) x 5 (h) x <5 (v)

    assert_allclose(row["sum_aper_area"], 15 * solid_angle_unit)  # 3 (w) x 5 (h)
    assert_allclose(row["mean"], 3.236648941040039 * flux_unit)
    assert_quantity_allclose(row["slice_wave"], 4.894499866699333 * u.um)


@pytest.mark.parametrize("cube_unit", [u.MJy / u.sr, u.MJy, u.MJy / PIX2])
def test_cubeviz_aperphot_cube_sr_and_pix2(cubeviz_helper,
                                           spectrum1d_cube_custom_fluxunit,
                                           cube_unit):
    # tests aperture photometry outputs between different inputs of flux (which
    # should be converted to a surface brighntess after loading), flux per sr
    # and flux per square pixel. the pixel area for per steradian cubes is
    # set so the output values between units will be the same

    cube = spectrum1d_cube_custom_fluxunit(fluxunit=cube_unit)
    cubeviz_helper.load_data(cube, data_label="test")

    aper = RectanglePixelRegion(center=PixCoord(x=3, y=1), width=1, height=1)
    bg = RectanglePixelRegion(center=PixCoord(x=2, y=0), width=1, height=1)
    cubeviz_helper.plugins['Subset Tools'].import_region(
        [aper, bg], combination_mode='new')

    plg = cubeviz_helper.plugins["Aperture Photometry"]._obj
    plg.dataset_selected = "test[FLUX]"
    plg.aperture_selected = "Subset 1"
    plg.background_selected = "Subset 2"

    #  Check that the default flux scaling is present for MJy / sr cubes
    if cube_unit == (u.MJy / u.sr):
        assert_allclose(plg.flux_scaling, 0.003631)
    else:
        assert_allclose(plg.flux_scaling, 0.0)

    # if cube is MJy / sr, set pixel area to 1 sr / pix2 so
    # we can directly compare outputs between per sr and per pixel cubes, which
    # will give the same results with this scaling
    if cube_unit == (u.MJy / u.sr):
        assert_allclose(plg.pixel_area, 0.01)  # check default
        plg.pixel_area = 1 * (u.sr).to(u.arcsec*u.arcsec)
        solid_angle_unit = u.sr

    else:
        # for per pixel cubes, set flux scaling to default for MJy / sr cubes
        # so we can directly compare. this shouldn't be populated automatically,
        # which is checked above
        plg.flux_scaling = 0.003631
        solid_angle_unit = PIX2
        cube_unit = u.MJy / solid_angle_unit  # cube unit in app is now per pix2

    plg.vue_do_aper_phot()
    row = cubeviz_helper.plugins['Aperture Photometry'].export_table()[0]

    # Basically, we should recover the input rectangle here, minus background.
    assert_allclose(row["xcenter"], 3 * u.pix)
    assert_allclose(row["ycenter"], 1 * u.pix)
    # (15 - 10) MJy/sr x 1 sr, will always be MJy since solid angle is multiplied out
    assert_allclose(row["sum"], 5.0 * u.MJy)

    assert_allclose(row["sum_aper_area"], 1 * PIX2)

    #  we forced area to be one sr so MJy / sr and MJy / pix2 gave the same result
    assert_allclose(row["pixarea_tot"], 1.0 * solid_angle_unit)

    # also forced flux_scaling to be the same for MJy / sr cubes, which get a
    # default value populated, and MJy / pix2 which have a default of 0.0
    assert_allclose(row["aperture_sum_mag"], -7.847359 * u.mag)

    assert_allclose(row["mean"], 5 * (cube_unit))
    # TODO: check if slice plugin has value_unit set correctly
    assert_quantity_allclose(row["slice_wave"], 0.46236 * u.um)


def test_cubeviz_aperphot_cube_orig_flux_mjysr(cubeviz_helper,
                                               spectrum1d_cube_custom_fluxunit):
    # this test is essentially the same as test_cubeviz_aperphot_cube_sr_and_pix2
    # but for a single surface brightness unit and without changing the pixel
    # area to make outputs the same. it was requested in review to keep both tests
    cube = spectrum1d_cube_custom_fluxunit(fluxunit=u.MJy / u.sr)
    cubeviz_helper.load_data(cube, data_label="test")

    aper = RectanglePixelRegion(center=PixCoord(x=3, y=1), width=1, height=1)
    bg = RectanglePixelRegion(center=PixCoord(x=2, y=0), width=1, height=1)
    cubeviz_helper.plugins['Subset Tools'].import_region([aper, bg],
                                                         combination_mode='new')

    plg = cubeviz_helper.plugins["Aperture Photometry"]._obj
    plg.dataset_selected = "test[FLUX]"
    plg.aperture_selected = "Subset 1"
    plg.background_selected = "Subset 2"

    # Make sure per steradian is handled properly.
    assert_allclose(plg.pixel_area, 0.01)
    assert_allclose(plg.flux_scaling, 0.003631)

    plg.vue_do_aper_phot()
    row = cubeviz_helper.plugins['Aperture Photometry'].export_table()[0]

    # Basically, we should recover the input rectangle here, minus background.
    assert_allclose(row["xcenter"], 3 * u.pix)
    assert_allclose(row["ycenter"], 1 * u.pix)
    assert_allclose(row["sum"], 1.1752215e-12 * u.MJy)  # (15 - 10) MJy/sr x 2.3504431e-13 sr
    assert_allclose(row["sum_aper_area"], 1 * PIX2)
    assert_allclose(row["pixarea_tot"], 2.350443053909789e-13 * u.sr)
    assert_allclose(row["aperture_sum_mag"], 23.72476627732448 * u.mag)
    assert_allclose(row["mean"], 5 * (u.MJy / u.sr))
    # TODO: check if slice plugin has value_unit set correctly
    assert_quantity_allclose(row["slice_wave"], 0.46236 * u.um)


def _compare_table_units(orig_tab, new_tab, orig_flux_unit=None,
                         new_flux_unit=None, equivalencies=None):

    # compare two photometry tables with different units row by row, and make
    # sure that the units are as expected, and that they are equivalent once
    # translated

    for i, row in enumerate(orig_tab):
        new_unit = new_tab[i]['unit'] or '-'
        orig_unit = row['unit'] or '-'
        if new_unit != '-' and orig_unit != '-':

            new_unit = u.Unit(new_unit)
            new = float(new_tab[i]['result']) * new_unit

            orig_unit = u.Unit(orig_unit)
            orig = float(row['result']) * orig_unit

            if 'var' in row['function']:  # variance is in units of flux/sb squared
                orig_converted = handle_squared_flux_unit_conversions(orig.value,
                                                                      orig_unit,
                                                                      new_unit,
                                                                      equivalencies)
            else:
                orig_converted = flux_conversion_general(orig.value,
                                                         orig_unit,
                                                         new_unit,
                                                         equivalencies)

            # low rtol for match, phot table is rounded
            assert_quantity_allclose(orig_converted, new, rtol=1e-03)


@pytest.mark.slow
@pytest.mark.parametrize("flux_unit", [u.Unit(x) for x in SPEC_PHOTON_FLUX_DENSITY_UNITS])
@pytest.mark.parametrize("angle_unit", [u.sr, PIX2])
@pytest.mark.parametrize("new_flux_unit", [u.Unit(x) for x in SPEC_PHOTON_FLUX_DENSITY_UNITS])
def test_cubeviz_aperphot_unit_conversions(cubeviz_helper,
                                           spectrum1d_cube_custom_fluxunit,
                                           flux_unit, angle_unit, new_flux_unit):
    """
    Test cubeviz aperture photometry with all possible unit conversions for
    cubes in spectral/photon surface brightness units (e.g. Jy/sr, Jy/pix2).

    The aperture photometry plugin should respect the choice of flux and angle
    unit selected in the Unit Conversion plugin, and inputs and results should
    be converted based on selection. All conversions between units in the
    flux dropdown menu in the unit conversion plugin should be supported
    by aperture photometry.
    """

    if new_flux_unit == flux_unit:  # skip 'converting' to same unit
        return

    cube_unit = flux_unit / angle_unit

    # get strings of input units
    flux_unit_str = flux_unit.to_string()
    angle_unit_str = angle_unit.to_string()
    cube_unit_str = cube_unit.to_string()
    new_flux_unit_str = new_flux_unit.to_string()

    # load cube with specified unit
    cube = spectrum1d_cube_custom_fluxunit(fluxunit=cube_unit, shape=(5, 5, 4),
                                           with_uncerts=True)
    cubeviz_helper.load_data(cube, data_label="test")

    # get plugins
    st = cubeviz_helper.plugins['Subset Tools']
    ap = cubeviz_helper.plugins['Aperture Photometry']._obj
    uc = cubeviz_helper.plugins['Unit Conversion']

    # load aperture
    aper = RectanglePixelRegion(center=PixCoord(x=2, y=3), width=1, height=1)
    st.import_region(aper, combination_mode='new')

    # select dataset and aperture in plugin
    ap.dataset_selected = "test[FLUX]"
    ap.aperture_selected = "Subset 1"

    # equivalencies for unit conversion, we only need u.spectral_density because
    # no flux<>sb conversions will occur in this plugin
    equiv = u.spectral_density(ap._cube_wave)

    # check initial unit traitlets are synced between ap. phot and unit conv. plugins
    assert uc.flux_unit.selected == ap.flux_scaling_display_unit == flux_unit_str
    assert uc.angle_unit.selected == angle_unit_str
    assert ap.display_unit == cube_unit_str
    assert ap.flux_scaling_display_unit == flux_unit_str

    # set background to manual and background/flux scaling to 1 to make it
    # easier to compare between unit conversions
    ap.background_selected == 'Manual'
    ap.background_value = 1.
    ap.flux_scaling = 1.

    # do aperture photometry with inital cube units to compare original results
    # to results after flux unit conversion
    ap.vue_do_aper_phot()
    orig_tab = Table(ap.results)

    # set to new unit
    uc.flux_unit.selected = new_flux_unit_str

    # make sure display units in aperture phot plugin reflect change
    assert (u.Unit(ap.display_unit) * angle_unit).to_string() == new_flux_unit
    assert ap.flux_scaling_display_unit == new_flux_unit_str

    # make sure background and flux scaling were converted to new unit
    assert_allclose((ap.background_value * new_flux_unit).to(flux_unit, equiv).value, 1.)
    assert_allclose((ap.flux_scaling * new_flux_unit).to(flux_unit, equiv).value, 1.)

    ap.vue_do_aper_phot()
    new_tab = Table(ap.results)

    # if ap. phot silently fails, then 'new_tab' will just be the last
    # calculated one, so make sure this didn't happen
    assert not np.all(orig_tab == new_tab)

    # compare output tables row by row between original and new unit
    _compare_table_units(orig_tab, new_tab, orig_flux_unit=flux_unit,
                         new_flux_unit=new_flux_unit_str,
                         equivalencies=equiv)

    # todo: figure out how to test radial profile and curve of growth plots
    # related ticket: JDAT 4962



import warnings
from traitlets import Bool, List, Unicode, observe

from jdaviz.core.template_mixin import PluginTemplateMixin, SelectPluginComponent, with_spinner
from jdaviz.core.registries import (loader_resolver_registry,
                                    loader_parser_registry,
                                    loader_importer_registry)
from jdaviz.core.user_api import LoaderUserApi

__all__ = ['BaseResolver', 'find_matching_resolver']


class FormatSelect(SelectPluginComponent):
    """
    Select component for the format field of a resolver.

    Parameters
    ----------
    plugin
        the parent plugin object
    items : str
        the name of the items traitlet defined in ``plugin``
    selected : str
        the name of the selected traitlet defined in ``plugin``
    default_mode : str, optional
        What mode to use when making the default selection.  Valid options: first, default_text,
        empty.
    """
    def __init__(self, plugin, items, selected, default_mode='first'):
        self._importers = {}
        super().__init__(plugin,
                         items=items,
                         selected=selected,
                         default_mode=default_mode)

    def _is_valid_item(self, item):
        return super()._is_valid_item(item, locals())

    @observe('filters')
    def _update_items(self, msg={}):
        if not self.plugin.is_valid:
            self.items = []
            self._apply_default_selection()
            return

        # check for valid parser > importer combinations given the current filters
        # and resolver inputs
        try:
            parser_input = self.plugin()
        except Exception:
            self.items = []
            self._apply_default_selection()
            return

        all_resolvers = []
        self._importers = {}
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            for parser_name, Parser in loader_parser_registry.members.items():
                this_parser = Parser(self.plugin.app, parser_input)
                try:
                    if this_parser.is_valid:
                        importer_input = this_parser()
                    else:
                        importer_input = None
                except Exception:
                    importer_input = None

                if importer_input is None:
                    continue
                for importer_name, Importer in loader_importer_registry.members.items():
                    this_importer = Importer(app=self.plugin.app, input=importer_input)
                    if this_importer.is_valid:
                        if self._is_valid_item(this_importer):
                            all_resolvers.append({'label': importer_name,
                                                  'parser': parser_name,
                                                  'importer': importer_name,
                                                  'target': this_importer.target})
                        # we'll store the importer even if it isn't valid according to the filters
                        # so that they can be used when compiling the list of target filters
                        self._importers[importer_name] = this_importer

        self.items = all_resolvers
        self._apply_default_selection()


class TargetSelect(SelectPluginComponent):
    """
    Select component for the target (filter on format) of a resolver.

    Parameters
    ----------
    plugin
        the parent plugin object
    items : str
        the name of the items traitlet defined in ``plugin``
    selected : str
        the name of the selected traitlet defined in ``plugin``
    default_mode : str, optional
        What mode to use when making the default selection.  Valid options: first, default_text,
        empty.
    """
    def __init__(self, plugin, items, selected, default_mode='first'):
        self._importers = {}
        super().__init__(plugin,
                         items=items,
                         selected=selected,
                         default_mode=default_mode)

    def _is_valid_item(self, item):
        return super()._is_valid_item(item, locals())

    def set_filter_target_in(self, targets):
        def filter_factory(targets):
            def filter_target(item):
                return item['label'] in targets
            return filter_target
        self.filters = [filter_factory(targets)]

    @observe('filters')
    def _update_items(self, msg={}):
        if not self.plugin.is_valid:
            self.items = []
            self._apply_default_selection()
            return

        # loop through choices on format and compile list of targets
        # note that the selection of a target may affect the available formats
        # so we want to store all importers in the target select even if they are not valid there
        # and use that list when compiling list of valid targets
        all_targets = list(set([importer.target for importer
                                in self.plugin.format._importers.values()]))

        all_items = [{'label': 'Any'}]+[{'label': target} for target in all_targets]
        self.items = [item for item in all_items if self._is_valid_item(item)]
        self._apply_default_selection()


class BaseResolver(PluginTemplateMixin):
    default_input = None
    requires_api_support = False

    importer_widget = Unicode().tag(sync=True)

    import_spinner = Bool(False).tag(sync=True)

    format_items_spinner = Bool(False).tag(sync=True)
    format_items = List().tag(sync=True)
    format_selected = Unicode().tag(sync=True)

    target_items = List().tag(sync=True)
    target_selected = Unicode().tag(sync=True)

    def __init__(self, *args, **kwargs):
        self.set_active_loader_callback = kwargs.pop('set_active_loader_callback', None)
        self.open_callback = kwargs.pop('open_callback', None)
        self.close_callback = kwargs.pop('close_callback', None)
        super().__init__(*args, **kwargs)

        # subclasses should call self._update_format_items on any change
        # to user-inputs that might affect the available formats
        self.format = FormatSelect(self,
                                   items='format_items',
                                   selected='format_selected')

        self.target = TargetSelect(self,
                                   items='target_items',
                                   selected='target_selected')

    @classmethod
    def from_input(cls, app, inp, **kwargs):
        self = cls(app=app)
        if self.default_input is None:
            raise NotImplementedError("Resolver subclass must implement default_input")  # noqa pragma: nocover
        setattr(self, self.default_input, inp)
        user_api = self.user_api
        for k, v in kwargs.items():
            if hasattr(user_api, k):
                setattr(user_api, k, v)
        return self

    @property
    def is_valid(self):
        # override by subclass
        return False  # pragma: nocover

    @property
    def input(self):
        if self.default_input is None:
            raise NotImplementedError("Resolver subclass must implement default_input")
        return getattr(self, self.default_input)

    def __call__(self):
        # override by subclass - must convert any inputs into something
        # that can be interpretted by at least one parser
        # (generally a filepath, file object, or python object)
        raise NotImplementedError("Resolver subclass must implement __call__")  # pragma: nocover

    @property
    def user_api(self):
        return LoaderUserApi(self)

    @with_spinner('format_items_spinner')
    def _update_format_items(self):
        self.format._update_items()
        self.target._update_items()  # assumes format._importers is updated from above

    @property
    def importer(self):
        # give access to the importer defined by the user-selection on format
        if not self.format.selected:
            raise ValueError("must select a format before accessing importer")
        return self.format._importers[self.format.selected]

    @observe('target_selected')
    def _on_target_selected_changed(self, change={}):
        def matches_target_factory(target):
            def matches_target(importer):
                return importer.target == target
            return matches_target

        if self.target_selected == 'Any':
            self.format.filters = []
        else:
            self.format.filters = [matches_target_factory(self.target_selected)]

    @observe('format_selected')
    def _on_format_selected_changed(self, change):
        if self.format_selected == '':
            self.importer_widget = ''
        else:
            self.importer_widget = "IPY_MODEL_" + self.importer.model_id

    def close_in_tray(self, close_sidebar=False):
        """
        Close the loader in the sidebar/tray.

        Parameters
        ----------
        close_sidebar : bool
            Whether to also close the sidebar itself.
        """
        if self.close_callback is not None:
            self.close_callback()
        if close_sidebar:
            self.app.state.drawer_content = ''

    def open_in_tray(self):
        """
        Show this resolver in the sidebar tray.
        """
        if self.set_active_loader_callback is None:
            raise NotImplementedError("set_active_loader_callback must be set to open dialog to specific tab")  # noqa
        self.set_active_loader_callback(self._registry_label)
        if self.open_callback is not None:
            self.open_callback()

    @with_spinner('import_spinner')
    def vue_import_clicked(self, *args, **kwargs):
        self.importer()


def find_matching_resolver(app, inp=None, resolver=None, format=None, target=None, **kwargs):
    valid_resolvers = []
    for resolver_name, Resolver in loader_resolver_registry.members.items():
        if resolver is not None and resolver != resolver_name:
            continue
        try:
            this_resolver = Resolver.from_input(app, inp, **kwargs)
        except Exception:
            this_resolver = None
        if this_resolver is None:
            continue
        try:
            is_valid = this_resolver.is_valid
        except Exception:
            is_valid = False
        if not is_valid:
            continue

        if target is not None:
            if target not in this_resolver.target.choices:
                continue
            this_resolver.target = target
        for fmt in this_resolver.format.choices:
            if format is not None and fmt != format:
                continue
            this_resolver.format.selected = fmt
            valid_resolvers.append((this_resolver, resolver_name, fmt))

    if len(valid_resolvers) == 0:
        raise ValueError("no valid resolvers found for input")
    elif len(valid_resolvers) > 1:
        vrs = [f"resolver={vr[1]} > format={vr[2]}" for vr in valid_resolvers]
        raise ValueError(f"multiple valid resolvers found for input: {vrs}")
    else:
        return valid_resolvers[0][0].user_api



import os

import pytest

from jdaviz.configs.default.plugins.export.export import HAS_OPENCV


# TODO: Remove skip when https://github.com/bqplot/bqplot/pull/1397/files#r726500097 is resolved.
@pytest.mark.skip(reason="Cannot test due to async JS callback")
# @pytest.mark.skipif(not HAS_OPENCV, reason="opencv-python is not installed")
def test_export_movie(cubeviz_helper, spectrum1d_cube, tmp_path):
    orig_path = os.getcwd()
    os.chdir(tmp_path)
    try:
        cubeviz_helper.load_data(spectrum1d_cube, data_label="test")
        plugin = cubeviz_helper.plugins["Export"]
        assert plugin._obj.i_start == 0
        assert plugin._obj.i_end == 1

        plugin.viewer_format = 'mp4'
        plugin._obj.export()
        assert os.path.isfile("mymovie.mp4"), tmp_path
    finally:
        os.chdir(orig_path)


@pytest.mark.skipif(HAS_OPENCV, reason="opencv-python is installed")
def test_no_opencv(cubeviz_helper, spectrum1d_cube):
    cubeviz_helper.load_data(spectrum1d_cube, data_label="test")
    plugin = cubeviz_helper.plugins["Export"]
    assert 'mp4' in plugin.viewer_format.choices
    assert not plugin._obj.movie_enabled
    plugin.viewer_format = 'mp4'
    with pytest.raises(ImportError, match="Please install opencv-python"):
        plugin.export()


@pytest.mark.skipif(not HAS_OPENCV, reason="opencv-python is not installed")
def test_export_movie_not_cubeviz(imviz_helper):
    plugin = imviz_helper.plugins["Export"]
    assert 'mp4' not in plugin.viewer_format.choices


@pytest.mark.skipif(not HAS_OPENCV, reason="opencv-python is not installed")
def test_export_movie_cubeviz_exceptions(cubeviz_helper, spectrum1d_cube):
    cubeviz_helper.load_data(spectrum1d_cube, data_label="test")
    cubeviz_helper.default_viewer._obj.shape = (100, 100)
    cubeviz_helper.app.get_viewer("uncert-viewer").shape = (100, 100)
    plugin = cubeviz_helper.plugins["Export"]
    assert plugin._obj.i_start == 0
    assert plugin._obj.i_end == 1

    plugin.viewer_format = 'mp4'
    plugin._obj.i_end = 0
    with pytest.raises(ValueError, match="No frames to write"):
        plugin.export()

    plugin._obj.i_end = 1
    plugin._obj.movie_fps = 0
    with pytest.raises(ValueError, match="Invalid frame rate"):
        plugin.export()

    plugin._obj.movie_fps = 5
    plugin.filename = "fake_path/mymovie.mp4"
    with pytest.raises(ValueError, match="Invalid path"):
        plugin.export()

    plugin.viewer = 'uncert-viewer'
    plugin.filename = ""
    with pytest.raises(ValueError, match="Invalid filename"):
        plugin.export()


@pytest.mark.skipif(not HAS_OPENCV, reason="opencv-python is not installed")
def test_export_movie_cubeviz_empty(cubeviz_helper):
    plugin = cubeviz_helper.plugins["Export"]
    assert plugin._obj.i_start == 0
    assert plugin._obj.i_end == 0

    plugin.viewer_format = 'mp4'
    with pytest.raises(ValueError, match="Selected viewer has no display shape"):
        plugin.export()


def test_export_plot_exceptions(cubeviz_helper):
    plugin = cubeviz_helper.plugins["Export"]

    plugin.filename = "/fake/path/image.png"
    with pytest.raises(ValueError, match="Invalid path"):
        plugin.export()



"""
This module is used to initialize spectral models to the data at hand.

This is used by model-fitting code that has to create spectral model
instances with sensible parameter values such that they can be used as
first guesses by the fitting algorithms.
"""
import numpy as np

import astropy.modeling.models as models
from astropy import units as u

from jdaviz.models import BlackBody

__all__ = [
    'initialize',
    'get_model_parameters'
]

AMPLITUDE = 'amplitude'
POSITION = 'position'
WIDTH = 'width'

MODELS = {
     'Const1D': models.Const1D,
     'Linear1D': models.Linear1D,
     'Polynomial1D': models.Polynomial1D,
     'Gaussian1D': models.Gaussian1D,
     'Voigt1D': models.Voigt1D,
     'Lorentz1D': models.Lorentz1D,
     'PowerLaw1D': models.PowerLaw1D,
     'BlackBody': BlackBody
     }


def get_model_parameters(model_cls, model_kwargs={}):
    if isinstance(model_cls, str):
        model_cls = MODELS.get(model_cls)

    if model_cls.__name__ == 'Polynomial1D':
        # then the parameters are not stored, as they depend on the polynomial order
        degree = model_kwargs.get('degree', 1)
        return [f'c{n}' for n in range(degree + 1)]

    return model_cls.param_names


def _get_model_name(model):
    class_string = str(model.__class__)
    return class_string.split('\'>')[0].split(".")[-1]


class _Linear1DInitializer(object):
    """
    Initialization that is specific to the Linear1D model.

    Notes
    -----
    In a way, we need this specialized initializer because
    the linear 1D model is more like a kind of polynomial.
    It doesn't mesh well with other non-linear models.
    """
    def initialize(self, instance, x, y):
        """
        Initialize the model

        Parameters
        ----------
        instance : `~astropy.modeling.Model`
            The model to initialize.

        x, y : numpy.ndarray
            The data to use to initialize from.

        Returns
        -------
        instance : `~astropy.modeling.Model`
            The initialized model.
        """
        if y.ndim == 3:
            # For cube fitting, need to collapse before this calculation
            y = np.nanmean(y, axis=(0, 1))
        slope, intercept = np.polynomial.Polynomial.fit(x.value.flatten(), y.value.flatten(), 1)

        instance.slope.value = slope
        instance.intercept.value = intercept

        return instance


class _WideBand1DInitializer(object):
    """
    Initialization that is applicable to all "wide band"
    models

    A "wide band" model is one that has an amplitude and
    a position in wavelength space where this amplitude
    is defined.

    Parameters
    ----------
    factor: float
        The scale factor to apply to the amplitude
    """
    def __init__(self, factor=1.0):
        self._factor = factor

    def initialize(self, instance, x, y):
        """
        Initialize the model

        Parameters
        ----------
        instance: `~astropy.modeling.Model`
            The model to initialize.

        x, y: numpy.ndarray
            The data to use to initialize from.

        Returns
        -------
        instance: `~astropy.modeling.Model`
            The initialized model.
        """
        y_mean = np.nanmean(y)
        x_range = x[-1] - x[0]
        position = x_range / 2.0 + x[0]

        name = _get_model_name(instance)

        _setattr(instance, name, AMPLITUDE, y_mean * self._factor)
        _setattr(instance, name, POSITION, position)

        return instance


class _LineProfile1DInitializer(object):
    """
    Initialization that is applicable to all "line profile"
    models.

    A "line profile" model is one that has an amplitude, a width,
    and a defined position in wavelength space.

    Parameters
    ----------
    factor: float
        The scale factor to apply to the amplitude
    """
    def __init__(self, factor=1.0):
        self._factor = factor

    def _set_width_attribute(self, instance, name, fwhm):
        """
        Each line profile class has its own way of naming
        and defining the width parameter. Subclasses should
        override this method to conform to the specific
        definitions.

        Parameters
        ----------
        name : str
            The attribute name

        instance: `~astropy.modeling.Model`
            The model to initialize.

        fwhm : float
            FWHM
        """
        raise NotImplementedError

    def initialize(self, instance, x, y):
        """
        Initialize the model

        Parameters
        ----------
        instance: `~astropy.modeling.Model`
            The model to initialize.

        x, y: numpy.ndarray
            The data to use to initialize from.

        Returns
        -------
        instance: `~astropy.modeling.Model`
            The initialized model.
        """

        if y.ndim == 3:
            # For cube fitting, need to collapse before these calculations
            y = np.nanmean(y, axis=(0, 1))

        # X centroid estimates the position
        centroid = np.sum(x * y) / np.sum(y)

        # width can be estimated by the weighted
        # 2nd moment of the X coordinate.
        dx = x - np.nanmean(x)
        fwhm = 2 * np.sqrt(np.sum((dx * dx) * y) / np.sum(y))

        # amplitude is derived from area.
        delta_x = x[1:] - x[:-1]
        sum_y = np.sum((y[1:] - np.min(y[1:])) * delta_x)
        height = sum_y / (fwhm / 2.355 * np.sqrt(2 * np.pi))

        name = _get_model_name(instance)

        _setattr(instance, name, AMPLITUDE, height * self._factor)
        _setattr(instance, name, POSITION, centroid)

        self._set_width_attribute(instance, name, fwhm)

        return instance


class _Width_LineProfile1DInitializer(_LineProfile1DInitializer):
    def _set_width_attribute(self, instance, name, fwhm):
        _setattr(instance, name, WIDTH, fwhm)


class _Sigma_LineProfile1DInitializer(_LineProfile1DInitializer):
    def _set_width_attribute(self, instance, name, fwhm):
        _setattr(instance, name, WIDTH, fwhm / 2.355)


class _BlackBodyInitializer:
    """
    Initialization that is specific to the BlackBody model.

    Notes
    -----
    The fit is sensitive to the scale parameter as if it ever tries
    to go below zero, the fit will get stuck at 0 (without imposed
    parameter limits)
    """
    def initialize(self, instance, x, y):
        """
        Initialize the model

        Parameters
        ----------
        instance: `~astropy.modeling.Model`
            The model to initialize.

        x, y: numpy.ndarray
            The data to use to initialize from.

        Returns
        -------
        instance: `~astropy.modeling.Model`
            The initialized model.
        """
        y_mean = np.nanmean(y)

        # The y-unit could contain a scale factor (like 1e-7 * FLAM).  We
        # need to account for this in our dimensionless scale factor guess.
        # We could make this smarter if we also made a guess for the temperature
        # based on the peak wavelength/frequency and then estimated the amplitude,
        # but just getting within an order of magnitude should help significantly
        y_unit_scaled = y.unit
        for native_output_unit in instance._native_output_units.values():
            if y_unit_scaled.is_equivalent(native_output_unit):
                y_mean = y_mean.to(native_output_unit)
                break

        instance.scale = y_mean.value * u.dimensionless_unscaled

        return instance


def _setattr(instance, mname, pname, value):
    """
    Sets parameter value by mapping parameter name to model type.

    Prevents the parameter value setting to be stopped on its tracks
    by non-existent model names or parameter names.

    Parameters
    ----------
    instance: `~astropy.modeling.Model`
        The model to initialize.

    mname: str
        Model name.

    pname: str
        Parameter name.

    value: any
        The value to assign.
    """
    try:
        setattr(instance, _p_names[mname][pname], value)
    except KeyError:
        pass


# This associates each initializer to its corresponding spectral model.
# Some models are not really line profiles, but their parameter names
# and roles are the same as in a typical line profile, so they can be
# initialized in the same way.
_initializers = {
    'Beta1D':                      _WideBand1DInitializer,  # noqa
    'Const1D':                     _WideBand1DInitializer,  # noqa
    'PowerLaw1D':                  _WideBand1DInitializer,  # noqa
    'BrokenPowerLaw1D':            _WideBand1DInitializer,  # noqa
    'ExponentialCutoffPowerLaw1D': _WideBand1DInitializer,  # noqa
    'LogParabola1D':               _WideBand1DInitializer,  # noqa
    'Box1D':                       _Width_LineProfile1DInitializer,  # noqa
    'Gaussian1D':                  _Sigma_LineProfile1DInitializer,  # noqa
    'Lorentz1D':                   _Width_LineProfile1DInitializer,  # noqa
    'Voigt1D':                     _Width_LineProfile1DInitializer,  # noqa
    'MexicanHat1D':                _Sigma_LineProfile1DInitializer,  # noqa
    'Trapezoid1D':                 _Width_LineProfile1DInitializer,  # noqa
    'Linear1D':                    _Linear1DInitializer,  # noqa
    'BlackBody':                   _BlackBodyInitializer,  # noqa
    # 'Spline1D':                   spline.Spline1DInitializer
}

# Models can have parameter names that are similar amongst them, but not quite the same.
# This maps the standard names used in the code to the actual names used by astropy.
_p_names = {
    'Gaussian1D':                  {AMPLITUDE: 'amplitude',  POSITION: 'mean', WIDTH: 'stddev'},  # noqa
    'GaussianAbsorption':          {AMPLITUDE: 'amplitude',  POSITION: 'mean', WIDTH: 'stddev'},  # noqa
    'Lorentz1D':                   {AMPLITUDE: 'amplitude',  POSITION: 'x_0',  WIDTH: 'fwhm'},  # noqa
    'Voigt1D':                     {AMPLITUDE: 'amplitude_L', POSITION: 'x_0',  WIDTH: 'fwhm_G'},  # noqa
    'Box1D':                       {AMPLITUDE: 'amplitude',  POSITION: 'x_0',  WIDTH: 'width'},  # noqa
    'MexicanHat1D':                {AMPLITUDE: 'amplitude',  POSITION: 'x_0',  WIDTH: 'sigma'},  # noqa
    'Trapezoid1D':                 {AMPLITUDE: 'amplitude',  POSITION: 'x_0',  WIDTH: 'width'},  # noqa
    'Beta1D':                      {AMPLITUDE: 'amplitude',  POSITION: 'x_0'},  # noqa
    'PowerLaw1D':                  {AMPLITUDE: 'amplitude',  POSITION: 'x_0'},  # noqa
    'ExponentialCutoffPowerLaw1D': {AMPLITUDE: 'amplitude',  POSITION: 'x_0'},  # noqa
    'LogParabola1D':               {AMPLITUDE: 'amplitude',  POSITION: 'x_0'},  # noqa
    'BrokenPowerLaw1D':            {AMPLITUDE: 'amplitude',  POSITION: 'x_break'},  # noqa
    'Const1D':                     {AMPLITUDE: 'amplitude'},  # noqa
    }


def initialize(instance, x, y):
    """
    Initialize given model.

    X and Y are for now Quantity arrays with the
    independent and dependent variables. It's assumed X values
    are stored in increasing order in the array.

    Parameters
    ----------
    instance : `~astropy.modeling.Model`
        The model to initialize.

    x, y : ndarray
        The data to use to initialize from.

    Returns
    -------
    instance : `~astropy.modeling.Model`
        The initialized model.
        If there are any errors, the instance is returned
        uninitialized.
    """
    if x is None or y is None:
        return instance

    name = _get_model_name(instance)

    try:
        initializer = _initializers[name]()

        return initializer.initialize(instance, x, y)

    except KeyError:
        return instance



# The code in this file has been forked from the ipyfilechooser package which
# was originally released under the license in licenses/IPYFILECHOOSER_LICENSE.rst
#
# This was forked to enable us to use our own buttons for the dialog and use
# this just for the file browser view, but we could switch back to using the
# main released version of ipyfilechooser if it was possible to toggle this via
# public API. However for now this is sufficient until we know whether we are
# going to try and implement a native file picker instead.


import fnmatch
import os
import string
import sys

from traitlets import Unicode

from ipywidgets import Dropdown, Select, Layout, GridBox, VBox, ValueWidget

__all__ = ['FileChooser']


def get_subpaths(path):
    """Walk a path and return a list of subpaths."""
    if os.path.isfile(path):
        path = os.path.dirname(path)

    paths = [path]
    path, tail = os.path.split(path)

    while tail:
        paths.append(path)
        path, tail = os.path.split(path)

    try:
        # Add Windows drive letters, but remove the current drive
        drives = get_drive_letters()
        drives.remove(paths[-1])
        paths.extend(drives)
    except ValueError:
        pass
    return paths


def has_parent(path):
    """Check if a path has a parent folder."""
    return os.path.basename(path) != ''


def match_item(item, filter_pattern):
    """Check if a string matches one or more fnmatch patterns."""
    if isinstance(filter_pattern, str):
        filter_pattern = [filter_pattern]

    idx = 0
    found = False

    while idx < len(filter_pattern) and not found:
        found |= fnmatch.fnmatch(item, filter_pattern[idx])
        idx += 1

    return found


def get_dir_contents(
        path,
        show_hidden=False,
        prepend_icons=False,
        filter_pattern=None):
    """Get directory contents."""
    files = list()
    dirs = list()

    if os.path.isdir(path):
        for item in os.listdir(path):
            append = True
            if item.startswith('.') and not show_hidden:
                append = False
            full_item = os.path.join(path, item)
            if append and os.path.isdir(full_item):
                dirs.append(item)
            elif append:
                if filter_pattern:
                    if match_item(item, filter_pattern):
                        files.append(item)
                else:
                    files.append(item)
        if has_parent(path):
            dirs.insert(0, '..')
    if prepend_icons:
        return prepend_dir_icons(sorted(dirs)) + sorted(files)
    else:
        return add_trailing_slash(sorted(dirs)) + sorted(files)


def add_trailing_slash(dir_list):
    return [dirname + '/' for dirname in dir_list]


def prepend_dir_icons(dir_list):
    """Prepend unicode folder icon to directory names."""
    return ['\U0001F4C1 ' + dirname for dirname in dir_list]


def get_drive_letters():
    """Get drive letters."""
    if sys.platform == 'win32':
        # Windows has drive letters
        return [
            '%s:\\' % d for d in string.ascii_uppercase
            if os.path.exists('%s:' % d)
        ]
    else:
        # Unix does not have drive letters
        return []


class FileChooser(VBox, ValueWidget):
    """FileChooser class."""

    _LBL_TEMPLATE = '<span style="margin-left:10px; color:{1};">{0}</span>'
    _LBL_NOFILE = 'No file selected'

    file_path = Unicode(allow_none=True)

    def __init__(
            self,
            path=os.getcwd(),
            filename='',
            show_hidden=False,
            use_dir_icons=False,
            filter_pattern=None,
            **kwargs):
        """Initialize FileChooser object."""
        self._default_path = path.rstrip(os.path.sep)
        self._default_filename = filename
        self._show_hidden = show_hidden
        self._use_dir_icons = use_dir_icons
        self._filter_pattern = filter_pattern

        # Widgets

        self._pathlist = Dropdown(
            description="",
            layout=Layout(
                width='auto',
                grid_area='pathlist'
            )
        )

        self._dircontent = Select(
            rows=8,
            layout=Layout(
                width='auto',
                grid_area='dircontent'
            )
        )

        # Widget observe handlers
        self._pathlist.observe(
            self._on_pathlist_select,
            names='value'
        )
        self._dircontent.observe(
            self._on_dircontent_select,
            names='value'
        )

        # Layout
        self._gb = GridBox(
            children=[
                self._pathlist,
                self._dircontent
            ],
            layout=Layout(
                width='500px',
                grid_gap='0px 0px',
                grid_template_rows='auto auto',
                grid_template_columns='60% 40%',
                grid_template_areas='''
                    'pathlist pathlist'
                    'dircontent dircontent'
                    '''
            )
        )

        # Call setter to set initial form values
        self._set_form_values(
            self._default_path,
            self._default_filename
        )

        self._initialize_form_values()

        # Call VBox super class __init__
        super().__init__(
            children=[self._gb],
            layout=Layout(width='auto'),
            **kwargs
        )

    def _set_form_values(self, path, filename):
        """Set the form values."""
        # Disable triggers to prevent selecting an entry in the Select
        # box from automatically triggering a new event.
        self._pathlist.unobserve(
            self._on_pathlist_select,
            names='value'
        )
        self._dircontent.unobserve(
            self._on_dircontent_select,
            names='value'
        )

        # Set form values
        self._pathlist.options = get_subpaths(path)
        self._pathlist.value = path

        # file/folder real names
        dircontent_real_names = get_dir_contents(
            path,
            show_hidden=self._show_hidden,
            prepend_icons=False,
            filter_pattern=self._filter_pattern
        )

        # file/folder display names
        dircontent_display_names = get_dir_contents(
            path,
            show_hidden=self._show_hidden,
            prepend_icons=self._use_dir_icons,
            filter_pattern=self._filter_pattern
        )

        # Dict to map real names to display names
        self._map_name_to_disp = {
            real_name: disp_name
            for real_name, disp_name in zip(
                dircontent_real_names,
                dircontent_display_names
            )
        }

        # Dict to map display names to real names
        self._map_disp_to_name = dict(
            reversed(item) for item in self._map_name_to_disp.items()
        )

        # Set _dircontent form value to display names
        self._dircontent.options = dircontent_display_names

        # If the value in the filename Text box equals a value in the
        # Select box and the entry is a file then select the entry.
        if ((filename in dircontent_real_names) and
                os.path.isfile(os.path.join(path, filename))):
            self._dircontent.value = self._map_name_to_disp[filename]
        else:
            self._dircontent.value = None

        # Re-enable triggers again
        self._pathlist.observe(
            self._on_pathlist_select,
            names='value'
        )
        self._dircontent.observe(
            self._on_dircontent_select,
            names='value'
        )

        self._update_file_path()

    def _on_pathlist_select(self, change):
        """Handle selecting a path entry."""
        self._set_form_values(
            change['new'],
            self._selected_filename
        )

    def _on_dircontent_select(self, change):
        """Handle selecting a folder entry."""
        new_path = os.path.realpath(
            os.path.join(
                self._selected_path,
                self._map_disp_to_name[change['new']]
            )
        )

        # Check if folder or file
        if os.path.isdir(new_path):
            path = new_path
            filename = None
        elif os.path.isfile(new_path):
            path = self._selected_path
            filename = self._map_disp_to_name[change['new']]

        self._set_form_values(
            path,
            filename
        )

    def _initialize_form_values(self):
        """Show the dialog."""

        # Show the form with the correct path and filename
        if ((self._selected_path is not None) and
                (self._selected_filename is not None)):
            path = self._selected_path
            filename = self._selected_filename
        else:
            path = self._default_path
            filename = self._default_filename

        self._set_form_values(path, filename)

    @property
    def _selected_path(self):
        return self._pathlist.value

    @property
    def _selected_filename(self):
        if self._dircontent.value is None:
            return None
        else:
            return self._map_disp_to_name[self._dircontent.value]

    def _update_file_path(self):
        if self._selected_filename is not None and self._selected_path is not None:
            self.file_path = os.path.join(self._selected_path,
                                          self._selected_filename)

    def refresh(self):
        """Re-render the form."""
        self._set_form_values(
            self._selected_path,
            self._selected_filename
        )

    @property
    def show_hidden(self):
        """Get _show_hidden value."""
        return self._show_hidden

    @show_hidden.setter
    def show_hidden(self, hidden):
        """Set _show_hidden value."""
        self._show_hidden = hidden
        self.refresh()

    @property
    def use_dir_icons(self):
        """Get _use_dir_icons value."""
        return self._use_dir_icons

    @use_dir_icons.setter
    def use_dir_icons(self, dir_icons):
        """Set _use_dir_icons value."""
        self._use_dir_icons = dir_icons
        self.refresh()

    @property
    def rows(self):
        """Get current number of rows."""
        return self._dircontent.rows

    @rows.setter
    def rows(self, rows):
        """Set number of rows."""
        self._dircontent.rows = rows

    @property
    def default_path(self):
        """Get the default_path value."""
        return self._default_path

    @default_path.setter
    def default_path(self, path):
        """Set the default_path."""
        self._default_path = path.rstrip(os.path.sep)
        self._set_form_values(
            self._default_path,
            self._selected_filename
        )

    @property
    def default_filename(self):
        """Get the default_filename value."""
        return self._default_filename

    @default_filename.setter
    def default_filename(self, filename):
        """Set the default_filename."""
        self._default_filename = filename
        self._set_form_values(
            self._selected_path,
            self._default_filename
        )

    @property
    def filter_pattern(self):
        """Get file name filter pattern."""
        return self._filter_pattern

    @filter_pattern.setter
    def filter_pattern(self, filter_pattern):
        """Set file name filter pattern."""
        self._filter_pattern = filter_pattern
        self.refresh()

    @property
    def selected(self):
        """Get selected value."""
        try:
            return os.path.join(
                self._selected_path,
                self._selected_filename
            )
        except TypeError:
            return None



import warnings

import astropy.units as u

__all__ = ['UserApiWrapper', 'PluginUserApi',
           'LoaderUserApi', 'ImporterUserApi',
           'ViewerUserApi']

_internal_attrs = ('_obj', '_expose', '_items', '_readonly', '_exclude_from_dict',
                   '__doc__', '_deprecation_msg', '_deprecated')


class UserApiWrapper:
    """
    This is an API wrapper around an internal object.  For a full list of attributes/methods,
    call dir(object).
    """
    def __init__(self, obj, expose=[], readonly=[], exclude_from_dict=[], deprecated=[]):
        self._obj = obj
        self._expose = list(expose) + list(readonly)
        self._readonly = readonly
        self._exclude_from_dict = exclude_from_dict
        self._deprecation_msg = None
        self._deprecated = deprecated
        if obj.__doc__ is not None:
            self.__doc__ = self.__doc__ + "\n\n\n" + obj.__doc__

    def __dir__(self):
        return self._expose

    def __repr__(self):
        return f'<{self._obj.__class__.__name__} API>'

    def __eq__(self, other):
        return self._obj.__eq__(other)

    def __getattr__(self, attr):
        if attr in _internal_attrs or attr not in self._expose:
            return super().__getattribute__(attr)

        if attr in self._deprecated:
            warnings.warn(f"{attr} is deprecated", DeprecationWarning)

        exp_obj = getattr(self._obj, attr)
        return getattr(exp_obj, 'user_api', exp_obj)

    def __setattr__(self, attr, value):
        if attr in _internal_attrs:
            return super().__setattr__(attr, value)
        if attr not in self._expose:
            raise ValueError(f"{attr} is not a valid attribute and cannot be set")

        if attr in self._readonly:
            raise AttributeError("cannot set read-only item")

        exp_obj = getattr(self._obj, attr)
        if hasattr(exp_obj, '__call__'):
            raise AttributeError(f"{attr} is a callable, cannot set to a value.  See help({attr}) for input arguments.")  # noqa
        from jdaviz.core.template_mixin import (SelectPluginComponent,
                                                UnitSelectPluginComponent,
                                                PlotOptionsSyncState,
                                                AddResults,
                                                AutoTextField)
        if isinstance(exp_obj, SelectPluginComponent):
            # this allows setting the selection directly without needing to access the underlying
            # .selected traitlet
            if isinstance(exp_obj, UnitSelectPluginComponent) and isinstance(value, u.Unit):
                value = value.to_string()
            exp_obj.selected = value
            return
        elif isinstance(exp_obj, AddResults):
            exp_obj.auto_label.value = value
            return
        elif isinstance(exp_obj, AutoTextField):
            exp_obj.value = value
            return
        elif isinstance(exp_obj, PlotOptionsSyncState):
            if not len(exp_obj.linked_states):
                raise ValueError("there are currently no synced glue states to set")

            # this allows setting the value immediately, and unmixing state, if appropriate,
            # even if the value matches the current value
            if value == exp_obj.value:
                exp_obj.unmix_state()
            else:
                # if there are choices, allow either passing the text or value
                text_to_value = {choice['text']: choice['value']
                                 for choice in exp_obj.sync.get('choices', [])}
                exp_obj.value = text_to_value.get(value, value)
            return

        return setattr(self._obj, attr, value)

    def _items(self):
        for attr in self._expose:
            try:
                yield attr, self.__getattr__(attr)
            except AttributeError:
                continue

    def to_dict(self):
        def _value(item):
            if hasattr(item, 'to_dict'):
                return _value(item.to_dict())
            if hasattr(item, 'selected'):
                return item.selected
            return item

        return {k: _value(getattr(self, k)) for k in self._expose
                if k not in ('api_hints_enabled', 'keep_active')
                and k not in self._exclude_from_dict
                and not hasattr(getattr(self, k), '__call__')}

    def from_dict(self, d):
        # loop through expose so that plugins can dictate the order that items should be populated
        for k in self._expose:
            if k not in d:
                continue
            v = d.get(k)
            if hasattr(getattr(self, k), '__call__'):
                raise ValueError(f"cannot overwrite callable {k}")
            if hasattr(getattr(self, k), 'from_dict') and isinstance(v, dict):
                getattr(self, k).from_dict(v)
            else:
                setattr(self, k, v)


class PluginUserApi(UserApiWrapper):
    """
    This is an API wrapper around an internal plugin.  For a full list of attributes/methods,
    call dir(plugin_object) and for help on any of those methods,
    call help(plugin_object.attribute).

    For example::
      help(plugin_object.show)
    """
    def __init__(self, plugin, expose=[], readonly=[], excl_from_dict=[], deprecated=[]):
        expose = list(set(list(expose) + ['open_in_tray', 'close_in_tray',
                                          'show']))
        if plugin.uses_active_status:
            expose += ['keep_active', 'as_active']
        self._deprecation_msg = None
        super().__init__(plugin, expose, readonly, excl_from_dict, deprecated)

    def __repr__(self):
        if self._deprecation_msg:
            warnings.warn(self._deprecation_msg, DeprecationWarning)
            super().__setattr__('_deprecation_msg', None)
        return f'<{self._obj._registry_label} API>'


class LoaderUserApi(UserApiWrapper):
    """
    This is an API wrapper around an internal loader/resolver.  For a full list of
    attributes/methods, call dir(loader_object) and for help on any of those methods,
    call ``help(loader_object.attribute)``.

    For example::

      help(loader_object.show)
    """
    def __init__(self, loader, expose=[], readonly=[], excl_from_dict=[], deprecated=[]):
        expose = list(set(list(expose) + ['format', 'target', 'importer', 'show',
                                          'open_in_tray', 'close_in_tray']))
        super().__init__(loader, expose, readonly, excl_from_dict, deprecated)

    def __repr__(self):
        return f'<{self._obj._registry_label} API>'


class ImporterUserApi(UserApiWrapper):
    """
    This is an API wrapper around an internal importer.  For a full list of attributes/methods,
    call dir(importer_object) and for help on any of those methods,
    call ``help(importer_object.attribute)``.

    For example::

      help(importer_object.show)
    """
    def __init__(self, importer, expose=[], readonly=[], excl_from_dict=[], deprecated=[]):
        expose = list(set(list(expose) + ['input', 'output', 'target', 'show']))
        if hasattr(importer, 'data_label'):
            expose += ['data_label']
        super().__init__(importer, expose, readonly, excl_from_dict, deprecated)

    def __call__(self):
        return self._obj()

    def __repr__(self):
        return f'<{self._obj._registry_label} API>'


class ViewerUserApi(UserApiWrapper):
    """
    This is an API wrapper around a viewer.  For a full list of attributes/methods,
    call dir(viewer_object) and for help on any of those methods,
    call help(viewer_object.attribute).

    For example::
      help(viewer_object.show)
    """
    def __init__(self, viewer, expose=[], readonly=[], excl_from_dict=[], deprecated=[]):
        expose = list(set(list(expose) + []))
        super().__init__(viewer, expose, readonly, excl_from_dict, deprecated)

    def __repr__(self):
        return f'<{self._obj.reference} API>'

    def __getattr__(self, *args, **kwargs):
        if super().__getattr__('_deprecation_msg'):
            warnings.warn(self._deprecation_msg, DeprecationWarning)
            super().__setattr__('_deprecation_msg', None)
        return super().__getattr__(*args, **kwargs)

    def __setattr__(self, *args, **kwargs):
        if hasattr(self, '_deprecation_msg') and self._deprecation_msg:
            warnings.warn(self._deprecation_msg, DeprecationWarning)
            super().__setattr__('_deprecation_msg', None)
        return super().__setattr__(*args, **kwargs)



# this module is based on solara/server/qt.py
import sys
from typing import List
import webbrowser
try:
    from qtpy.QtWidgets import QApplication
    from qtpy.QtWebEngineWidgets import QWebEngineView
    from qtpy.QtWebChannel import QWebChannel
    from qtpy import QtCore, QtGui
except ModuleNotFoundError as e:
    raise ModuleNotFoundError("""Qt browser requires Qt dependencies, run:
$ pip install jdaviz[qt]
to install.""") from e
import signal
from pathlib import Path

HERE = Path(__file__).parent


# setUrlRequestInterceptor, navigationRequested and acceptNavigationRequest
# all trigger the websocket to disconnect, so we need to block cross origin
# requests on the frontend/browser side by intercepting clicks on links

cross_origin_block_js = """
var script = document.createElement('script');
script.src = 'qrc:///qtwebchannel/qwebchannel.js';
document.head.appendChild(script);
script.onload = function() {
    new QWebChannel(qt.webChannelTransport, function(channel) {
        let py_callback = channel.objects.py_callback;

        document.addEventListener('click', function(event) {
            let target = event.target;
            while (target && target.tagName !== 'A') {
                target = target.parentNode;
            }

            if (target && target.tagName === 'A') {
                const linkOrigin = new URL(target.href).origin;
                const currentOrigin = window.location.origin;

                if (linkOrigin !== currentOrigin) {
                    event.preventDefault();
                    console.log("Blocked cross-origin navigation to:", target.href);
                    py_callback.open_link(target.href);  // Call Python method
                }
            }
        }, true);
    });
};
"""


class PyCallback(QtCore.QObject):
    @QtCore.Slot(str)
    def open_link(self, url):
        webbrowser.open(url)


class QWebEngineViewWithPopup(QWebEngineView):
    # keep a strong reference to all windows
    windows: List = []

    def __init__(self):
        super().__init__()
        self.page().newWindowRequested.connect(self.handle_new_window_request)

        # Set up WebChannel and py_callback object
        self.py_callback = PyCallback()
        self.channel = QWebChannel()
        self.channel.registerObject("py_callback", self.py_callback)
        self.page().setWebChannel(self.channel)

        self.loadFinished.connect(self._inject_javascript)

    def _inject_javascript(self, ok):
        self.page().runJavaScript(cross_origin_block_js)

    def handle_new_window_request(self, info):
        webview = QWebEngineViewWithPopup()
        geometry = info.requestedGeometry()
        width = geometry.width()
        parent_size = self.size()
        if width == 0:
            width = parent_size.width()
        height = geometry.height()
        if height == 0:
            height = parent_size.height()
        print("new window", info.requestedUrl(), width, height)
        webview.resize(width, height)
        webview.setUrl(info.requestedUrl())
        webview.show()
        QWebEngineViewWithPopup.windows.append(webview)
        return webview


def run_qt(url, app_name="Jdaviz"):
    app = QApplication([])
    web = QWebEngineViewWithPopup()
    web.setUrl(QtCore.QUrl(url))
    web.resize(1024, 1024)
    web.show()

    app.setApplicationDisplayName(app_name)
    app.setApplicationName(app_name)
    web.setWindowTitle(app_name)
    app.setWindowIcon(QtGui.QIcon(str(HERE / "data/icons/imviz_icon.svg")))
    if sys.platform.startswith("darwin"):
        # Set app name, if PyObjC is installed
        # Python 2 has PyObjC preinstalled
        # Python 3: pip3 install pyobjc-framework-Cocoa
        try:
            from Foundation import NSBundle

            bundle = NSBundle.mainBundle()
            if bundle:
                app_info = bundle.localizedInfoDictionary() or bundle.infoDictionary()
                if app_info is not None:
                    app_info["CFBundleName"] = app_name
                    app_info["CFBundleDisplayName"] = app_name
        except ModuleNotFoundError:
            pass

    # without this, ctrl-c does not work in the terminal
    signal.signal(signal.SIGINT, signal.SIG_DFL)
    app.exec_()



import os
import re

import numpy as np
import pytest
from astropy import units as u
from astropy.io import fits
from astropy.nddata import NDData
from glue.core.roi import CircularROI
from regions import Regions, CircleSkyRegion
from specutils import Spectrum1D, SpectralRegion
from pathlib import Path


@pytest.mark.usefixtures('_jail')
class TestExportSubsets:
    """
    Tests for exporting subsets. Currently limited to non-composite spatial
    subsets.
    """

    def test_basic_export_subsets_imviz(self, imviz_helper):

        data = NDData(np.ones((500, 500)) * u.nJy)

        imviz_helper.load_data(data)
        subset_plugin = imviz_helper.plugins['Subset Tools']
        subset_plugin.import_region(CircularROI(xc=250, yc=250, radius=100))

        export_plugin = imviz_helper.plugins['Export']._obj
        export_plugin.subset.selected = 'Subset 1'

        assert export_plugin.subset_format.selected == 'fits'  # default format
        assert export_plugin.subset_invalid_msg == ''  # for non-composite spatial

        assert export_plugin.filename.value.endswith('.fits')
        export_plugin.export()
        assert os.path.isfile(export_plugin.filename.value)

        # read exported file back in
        with fits.open(export_plugin.filename.value) as hdu:
            fits_region = hdu[1].data[0]

        assert fits_region[0] == 'circle'
        assert fits_region[1] == fits_region[2] == 250.0
        assert fits_region[3] == 100.0
        assert fits_region[4] == 0.0

        # now test changing file format
        export_plugin.subset_format.selected = 'reg'
        assert export_plugin.filename.value.endswith('.reg')
        export_plugin.export()
        assert os.path.isfile(export_plugin.filename.value)

        # read exported file back in
        region = Regions.read(export_plugin.filename.value)[0]
        assert region.center.x == 250.0
        assert region.center.y == 250.0
        assert region.radius == 100.0

        # changing file name
        export_plugin.filename_value = 'test'
        export_plugin.export()
        assert os.path.isfile('test.reg')

        # test that invalid file extension raises an error
        with pytest.raises(ValueError,
                           match=re.escape("'x' not one of ['fits', 'reg', 'ecsv'], reverting selection to 'reg'")):  # noqa
            export_plugin.subset_format.selected = 'x'

    def test_not_implemented(self, cubeviz_helper, spectral_cube_wcs):
        """
        Test that trying to export non-supported subsets
        (spectral and composite) produces
        the correct warning message to display in UI).
        """

        data = Spectrum1D(flux=np.ones((500, 500, 2)) * u.nJy,
                          wcs=spectral_cube_wcs)
        cubeviz_helper.load_data(data)
        subset_plugin = cubeviz_helper.plugins['Subset Tools']
        subset_plugin.import_region(CircularROI(xc=255, yc=255, radius=50))
        subset_plugin.combination_mode = 'and'
        subset_plugin.import_region(CircularROI(xc=200, yc=250, radius=50))

        export_plugin = cubeviz_helper.plugins['Export']._obj
        export_plugin.subset.selected = 'Subset 1'

        assert export_plugin.subset_invalid_msg == 'Export for composite subsets not yet supported.'

    def test_export_subsets_wcs(self, imviz_helper, spectral_cube_wcs):

        # using cube WCS instead of 2d imaging wcs for consistancy with
        # cubeviz test. accessing just the spatial part of this.
        wcs = spectral_cube_wcs.celestial

        data = NDData(np.ones((500, 500)) * u.nJy, wcs=wcs)

        imviz_helper.load_data(data)  # load data twice so we can link them
        imviz_helper.load_data(data)

        imviz_helper.link_data(align_by='wcs')
        subset_plugin = imviz_helper.plugins['Subset Tools']
        subset_plugin.import_region(CircularROI(xc=8, yc=6, radius=.2))

        export_plugin = imviz_helper.plugins['Export']._obj
        export_plugin.subset.selected = 'Subset 1'

        assert export_plugin.subset_invalid_msg == ''  # for non-composite spatial

        # test changing link type results in an output file with a sky region
        export_plugin.subset_format.selected = 'reg'
        export_plugin.filename_value = 'sky_region.reg'
        export_plugin.export()

        assert os.path.isfile('sky_region.reg')

        assert isinstance(Regions.read('sky_region.reg')[0], CircleSkyRegion)

    def test_basic_export_subsets_cubeviz(self, cubeviz_helper, spectral_cube_wcs):

        data = Spectrum1D(flux=np.ones((128, 128, 256)) * u.nJy, wcs=spectral_cube_wcs)

        cubeviz_helper.load_data(data)
        subset_plugin = cubeviz_helper.plugins['Subset Tools']
        subset_plugin.import_region(CircularROI(xc=50, yc=50, radius=10))

        export_plugin = cubeviz_helper.plugins['Export']._obj
        export_plugin.subset.selected = 'Subset 1'

        assert export_plugin.subset_format.selected == 'fits'  # default format

        assert export_plugin.filename.value.endswith('.fits')
        export_plugin.export()
        assert os.path.isfile(export_plugin.filename.value)

        # read exported file back in
        with fits.open(export_plugin.filename.value) as hdu:
            fits_region = hdu[1].data[0]

        assert fits_region[0] == 'circle'
        assert fits_region[1] == fits_region[2] == 50.0
        assert fits_region[3] == 10.0
        assert fits_region[4] == 0.0

        # now test changing file format
        export_plugin.subset_format.selected = 'reg'
        export_plugin.export()
        assert os.path.isfile(export_plugin.filename.value)

        # read exported file back in
        region = Regions.read(export_plugin.filename.value)[0]
        assert region.center.x == 50.0
        assert region.center.y == 50.0
        assert region.radius == 10.0

        # changing file name
        export_plugin.filename_value = 'test'
        export_plugin.export()
        assert os.path.isfile('test.reg')

        # Overwrite not enable, so no-op with warning.
        export_plugin.export(raise_error_for_overwrite=False)
        assert export_plugin.overwrite_warn

        # Changing filename should clear warning.
        old_filename = export_plugin.filename_value
        export_plugin.filename_value = "foo"
        assert not export_plugin.overwrite_warn
        export_plugin.filename_value = old_filename

        # Overwrite not enable, but with exception from API by default.
        with pytest.raises(FileExistsError, match=".* exists but overwrite=False"):
            export_plugin.export()
        assert export_plugin.overwrite_warn

        # User forces overwrite.
        export_plugin.export(overwrite=True)
        assert not export_plugin.overwrite_warn

        # test that invalid file extension raises an error
        with pytest.raises(ValueError,
                           match=re.escape("'x' not one of ['fits', 'reg', 'ecsv'], reverting selection to 'reg'")):  # noqa
            export_plugin.subset_format.selected = 'x'

        # Test that selecting disabled option raises an error
        with pytest.raises(ValueError, match="Cannot export Subset 1 in ecsv format, reverting selection to fits"):  # noqa
            export_plugin.subset_format.selected = 'ecsv'

        # test that attempting to save a composite subset raises an error
        subset_plugin.combination_mode.selected = 'and'
        subset_plugin.import_region(CircularROI(xc=25, yc=25, radius=5))
        subset_plugin.import_region(CircularROI(xc=20, yc=25, radius=5))

        with pytest.raises(NotImplementedError,
                           match='Subset can not be exported - Export for composite subsets not yet supported.'):  # noqa
            export_plugin.export()

        # Test saving spectral subset
        subset_plugin.combination_mode = 'new'
        spectral_axis_unit = u.Unit(
            cubeviz_helper.plugins['Unit Conversion'].spectral_unit.selected)
        subset_plugin.import_region(SpectralRegion(5 * spectral_axis_unit,
                                                   15.5 * spectral_axis_unit))
        export_plugin.subset.selected = 'Subset 2'

        # Format should auto-update to first non-disabled entry
        assert export_plugin.subset_format.selected == 'ecsv'
        for format in export_plugin.subset_format.items:
            if format['label'] != 'ecsv':
                assert format['disabled']
            else:
                assert format['disabled'] is False

        export_plugin.filename_value = "test_spectral_region"
        export_plugin.export()
        assert os.path.isfile('test_spectral_region.ecsv')


@pytest.mark.usefixtures('_jail')
def test_export_cubeviz_spectrum_viewer(cubeviz_helper, spectrum1d_cube):
    cubeviz_helper.load_data(spectrum1d_cube, data_label='test')

    ep = cubeviz_helper.plugins["Export"]
    ep.viewer = 'spectrum-viewer'
    ep.viewer_format = 'png'
    ep.export()

    ep.viewer_format = 'svg'
    ep.export()


@pytest.mark.usefixtures('_jail')
def test_export_data(cubeviz_helper, spectrum1d_cube):
    cubeviz_helper.load_data(spectrum1d_cube, data_label='test')
    mm = cubeviz_helper.plugins["Moment Maps"]
    mm.dataset = 'test[FLUX]'
    mm._obj.dataset_selected = 'test[FLUX]'
    mm.n_moment = 0
    mm.calculate_moment()
    assert mm._obj.results_label == 'moment 0'
    cubeviz_helper.app.add_data_to_viewer(
        cubeviz_helper._default_flux_viewer_reference_name, 'moment 0'
    )
    ep = cubeviz_helper.plugins["Export"]._obj
    assert 'test[FLUX]' not in ep.dataset.choices

    ep.dataset_selected = 'moment 0'
    assert ep.dataset_format.selected == 'fits'
    ep.export()
    assert os.path.isfile(ep.filename.value)
    assert ep.data_invalid_msg == ''


def test_disable_export_for_unsupported_units(specviz2d_helper):
    dn_per_s = u.DN / u.s
    data = np.zeros((5, 10))
    data[3] = np.arange(10)
    data = Spectrum1D(flux=data*dn_per_s, spectral_axis=data[3]*u.um)
    specviz2d_helper.load_data(data)

    gs = specviz2d_helper.plugins["Gaussian Smooth"]
    smooth_source_dataset = "Spectrum 1D"
    gs.dataset = smooth_source_dataset
    gs.stddev = 3
    gs.smooth(add_data=True)

    ep = specviz2d_helper.plugins["Export"]._obj
    assert "Spectrum 1D smooth stddev-3.0" in ep.dataset.choices
    ep.dataset_selected = "Spectrum 1D smooth stddev-3.0"
    assert ep.dataset.selected_obj.unit == "DN/s"
    assert ep.data_invalid_msg == "Export Disabled: The unit DN / s could not be saved in native FITS format."  # noqa


class TestExportPluginPlots:

    def test_basic_export_plugin_plots(self, imviz_helper):
        """
        Test basic funcionality of exporting plugin plots
        from the export plugin. Tests on the 'Plot Options: stretch_hist'
        plot, which exists upon loading data, and also that plots that
        may have been initialized but are empty are not displayed in
        the Export plugin.
        """
        data = NDData(np.ones((500, 500)) * u.nJy)

        imviz_helper.load_data(data)

        export_plugin = imviz_helper.plugins['Export']._obj
        export_plugin.plugin_plot.selected = 'Plot Options: stretch_hist'

        assert export_plugin.plugin_plot_format.selected == 'png'  # should be default format
        # and change file type
        export_plugin.plugin_plot_format.selected = 'svg'

        assert export_plugin.filename.value.endswith('.svg')
        # change filename
        export_plugin.filename_value = 'test_export_plugin_plot'

        # just check that it doesn't crash, since we can't download
        export_plugin.export()

        # make sure that the only valid option for export is this plugin,
        # not the other plots that exist but are empty (ap phot and line profile)
        # this might change down the line if new plots are added.
        available_plots = [x['label'] for x in export_plugin.plugin_plot.items]
        assert len(available_plots) == 1
        assert available_plots[0] == 'Plot Options: stretch_hist'

    def test_ap_phot_plot_export(self, imviz_helper):
        """
        Test export functionality for plot from the aperture photometry
        plugin.
        """

        data = NDData(np.ones((500, 500)) * u.nJy)

        imviz_helper.load_data(data)

        export_plugin = imviz_helper.plugins['Export']._obj
        subset_plugin = imviz_helper.plugins['Subset Tools']
        subset_plugin.import_region(CircularROI(xc=250, yc=250, radius=100))

        phot_plugin = imviz_helper.app.get_tray_item_from_name('imviz-aper-phot-simple')
        phot_plugin.aperture_selected = 'Subset 1'

        phot_plugin.vue_do_aper_phot()
        assert phot_plugin.plot_available

        available_plots = [x['label'] for x in export_plugin.plugin_plot.items]
        assert 'Aperture Photometry: plot' in available_plots

        export_plugin.plugin_plot.selected = 'Aperture Photometry: plot'

        # change filename
        export_plugin.filename_value = 'test_export_plugin_plot'
        # and change file type
        export_plugin.plugin_plot_format.selected = 'svg'

        # just check that it doesn't crash, since we can't download
        export_plugin.export()

    def test_figure_export(self, imviz_helper):

        data = NDData(np.ones((500, 500)) * u.nJy)

        imviz_helper.load_data(data)

        export_plugin = imviz_helper.plugins['Export']._obj

        export_plugin.export(filename=None)

        # attempt to save a figure back to back
        try:
            export_plugin.export(filename='img.png')
        except ValueError as e:
            assert str(e) == "previous png export is still in progress. Wait to complete before making another call to save_figure"  # noqa: E501

    def test_filepath_convention(self, imviz_helper):
        data = NDData(np.ones((500, 500)) * u.nJy)
        imviz_helper.load_data(data)
        export_plugin = imviz_helper.plugins['Export']._obj

        # Set filename value using OS-independent Path methods
        export_plugin.filename_value = str(Path('/') / 'img.png')
        assert os.path.abspath(export_plugin.default_filepath) == os.path.abspath(export_plugin.filename_value)  # noqa: E501

        export_plugin.filename_value = str(Path('~') / 'img.png')
        expected_path = str(Path('~').expanduser() / 'img.png')
        assert export_plugin.default_filepath == expected_path

        export_plugin.filename_value = str(Path('..') / 'img.png')
        expected_path = str((Path('..') / 'img.png').resolve())
        assert export_plugin.default_filepath == expected_path

        export_plugin.filename_value = str(Path('.') / 'img.png')
        expected_path = str((Path('.') / 'img.png').resolve())
        assert export_plugin.default_filepath == expected_path



from PyInstaller.utils.hooks import collect_data_files, copy_metadata, collect_submodules

hiddenimports = collect_submodules("astroquery")
datas = collect_data_files('astroquery')
datas += copy_metadata('astroquery')



import warnings

import numpy as np
import pytest
from astropy import units as u
from astropy.io import fits
from astropy.modeling import models
from astropy.nddata import StdDevUncertainty
from astropy.tests.helper import assert_quantity_allclose
from astropy.wcs import WCS
from glue.core.roi import XRangeROI
from numpy.testing import assert_allclose, assert_array_equal
from specutils.spectra import Spectrum1D

from jdaviz.configs.default.plugins.model_fitting import fitting_backend as fb
from jdaviz.configs.default.plugins.model_fitting import initializers
from jdaviz.core.custom_units_and_equivs import PIX2
from jdaviz.conftest import _create_spectrum1d_cube_with_fluxunit

SPECTRUM_SIZE = 200  # length of spectrum


def build_spectrum(sigma=0.1):
    g1 = models.Gaussian1D(1, 4.6, 0.2)
    g2 = models.Gaussian1D(2.5, 5.5, 0.1)
    g3 = models.Gaussian1D(-1.7, 8.2, 0.1)

    x = np.linspace(0, 10, SPECTRUM_SIZE)
    y = g1(x) + g2(x) + g3(x)

    noise = np.random.normal(4., sigma, x.shape)

    return x, y + noise


def test_model_params():
    # expected model parameters:
    model_parameters = {"Gaussian1D": ["amplitude", "stddev", "mean"],
                        "Const1D": ["amplitude"],
                        "Linear1D": ["slope", "intercept"],
                        "Polynomial1D": ["c0", "c1"],
                        "PowerLaw1D": ["amplitude", "x_0", "alpha"],
                        "Lorentz1D": ["amplitude", "x_0", "fwhm"],
                        "Voigt1D": ["x_0", "amplitude_L", "fwhm_L", "fwhm_G"],
                        "BlackBody": ["temperature", "scale"],
                        }

    for model_name in initializers.MODELS.keys():
        if model_name not in model_parameters.keys():
            # this would be caught later by the assertion anyways,
            # but raising an error will be more clear that the
            # test needs to be updated rather than the code breaking
            raise ValueError(f"{model_name} not in test dictionary of expected parameters")
        expected_params = model_parameters.get(model_name, [])
        params = initializers.get_model_parameters(model_name)
        assert len(params) == len(expected_params)
        assert np.all([p in expected_params for p in params])


def test_model_ids(cubeviz_helper, spectral_cube_wcs):
    cubeviz_helper.load_data(Spectrum1D(flux=np.ones((3, 4, 5)) * u.nJy, wcs=spectral_cube_wcs),
                             data_label='test')
    plugin = cubeviz_helper.plugins["Model Fitting"]._obj
    plugin.dataset_selected = 'Spectrum (sum)'
    plugin.component_models = [{'id': 'valid_string_already_exists'}]
    plugin.model_comp_selected = 'Linear1D'

    with pytest.raises(
            ValueError,
            match="model component label 'valid_string_already_exists' already in use"):
        plugin.comp_label = 'valid_string_already_exists'
        plugin.vue_add_model({})

    with pytest.raises(
            ValueError,
            match="invalid model component label 'invalid-string'"):
        plugin.comp_label = 'invalid-string'
        plugin.vue_add_model({})


def test_parameter_retrieval(cubeviz_helper, spectral_cube_wcs):

    flux_unit = u.nJy
    sb_unit = flux_unit / PIX2
    wav_unit = u.Hz

    flux = np.ones((3, 4, 5))
    flux[2, 2, :] = [1, 2, 3, 4, 5]
    cubeviz_helper.load_data(Spectrum1D(flux=flux * flux_unit, wcs=spectral_cube_wcs),
                             data_label='test')

    plugin = cubeviz_helper.plugins["Model Fitting"]

    # since cube_fit is true, output fit should be in SB units
    # even though the spectral y axis is in 'flux' by default
    plugin.cube_fit = True

    assert cubeviz_helper.app._get_display_unit('spectral') == wav_unit
    assert cubeviz_helper.app._get_display_unit('spectral_y') == flux_unit
    assert cubeviz_helper.app._get_display_unit('sb') == sb_unit

    plugin.create_model_component("Linear1D", "L")
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', message='Model is linear in parameters.*')
        plugin.calculate_fit()

    params = cubeviz_helper.plugins['Model Fitting'].get_model_parameters()
    slope_res = np.zeros((3, 4))
    slope_res[2, 2] = 1.0

    slope_res = slope_res * sb_unit / wav_unit
    intercept_res = np.ones((3, 4))
    intercept_res[2, 2] = 0
    intercept_res = intercept_res * sb_unit
    assert_quantity_allclose(params['model']['slope'], slope_res,
                             atol=1e-10 * sb_unit / wav_unit)
    assert_quantity_allclose(params['model']['intercept'], intercept_res,
                             atol=1e-10 * sb_unit)


@pytest.mark.parametrize('unc', ('zeros', None))
def test_fitting_backend(unc):
    np.random.seed(42)

    x, y = build_spectrum()

    # The uncertainty array of all zero should be ignored when fitting
    if unc == "zeros":
        uncertainties = StdDevUncertainty(np.zeros(y.shape)*u.Jy)
    elif unc is None:
        uncertainties = None
    spectrum = Spectrum1D(flux=y*u.Jy, spectral_axis=x*u.um, uncertainty=uncertainties)

    g1f = models.Gaussian1D(0.7*u.Jy, 4.65*u.um, 0.3*u.um, name='g1')
    g2f = models.Gaussian1D(2.0*u.Jy, 5.55*u.um, 0.3*u.um, name='g2')
    g3f = models.Gaussian1D(-2.*u.Jy, 8.15*u.um, 0.2*u.um, name='g3')
    zero_level = models.Const1D(1.*u.Jy, name='const1d')

    model_list = [g1f, g2f, g3f, zero_level]
    expression = "g1 + g2 + g3 + const1d"

    # Returns the initial model
    fm, fitted_spectrum = fb.fit_model_to_spectrum(spectrum, model_list, expression,
                                                   run_fitter=False)

    parameters_expected = np.array([0.7, 4.65, 0.3, 2., 5.55, 0.3, -2.,
                                    8.15, 0.2, 1.])
    assert_allclose(fm.parameters, parameters_expected, atol=1e-5)

    # Returns the fitted model
    fm, fitted_spectrum = fb.fit_model_to_spectrum(spectrum, model_list, expression,
                                                   run_fitter=True)

    parameters_expected = np.array([1.0104705, 4.58956282, 0.19590464, 2.39892026,
                                    5.49867754, 0.10834472, -1.66902953, 8.19714439,
                                    0.09535613, 3.99125545])
    assert_allclose(fm.parameters, parameters_expected, atol=1e-5)


@pytest.mark.parametrize('unc', ('zeros', None))
def test_cube_fitting_backend(cubeviz_helper, unc, tmp_path):
    np.random.seed(42)

    SIGMA = 0.1  # noise in data
    TOL = 0.4  # test tolerance
    IMAGE_SIZE_X = 15
    IMAGE_SIZE_Y = 14

    # Flux cube oriented as in JWST data. To build a Spectrum1D
    # instance with this, one need to transpose it so the spectral
    # axis direction corresponds to the last index.
    flux_cube = np.zeros((SPECTRUM_SIZE, IMAGE_SIZE_X, IMAGE_SIZE_Y))

    # Generate list of all spaxels to be fitted
    _spx = [[(x, y) for x in range(IMAGE_SIZE_X)] for y in range(IMAGE_SIZE_Y)]
    spaxels = [item for sublist in _spx for item in sublist]

    # Fill cube spaxels with spectra that differ from
    # each other only by their noise component.
    x, _ = build_spectrum()
    for spx in spaxels:
        flux_cube[:, spx[0], spx[1]] = build_spectrum(sigma=SIGMA)[1]

    # Transpose so it can be packed in a Spectrum1D instance.
    flux_cube = flux_cube.transpose(1, 2, 0)  # (15, 14, 200)
    cube_wcs = WCS({
        'WCSAXES': 3, 'RADESYS': 'ICRS', 'EQUINOX': 2000.0,
        'CRPIX3': 38.0, 'CRPIX2': 38.0, 'CRPIX1': 1.0,
        'CRVAL3': 205.4384, 'CRVAL2': 27.004754, 'CRVAL1': 0.0,
        'CDELT3': 0.01, 'CDELT2': 0.01, 'CDELT1': 0.05,
        'CUNIT3': 'deg', 'CUNIT2': 'deg', 'CUNIT1': 'um',
        'CTYPE3': 'RA---TAN', 'CTYPE2': 'DEC--TAN', 'CTYPE1': 'WAVE'})

    # Mask part of the spectral axis to later ensure that it gets propagated through:
    mask = np.zeros_like(flux_cube).astype(bool)
    mask[..., :SPECTRUM_SIZE // 10] = True

    # The uncertainty array of all zero should be ignored when fitting
    if unc == "zeros":
        uncertainties = StdDevUncertainty(np.zeros(flux_cube.shape)*u.Jy)
    elif unc is None:
        uncertainties = None

    spectrum = Spectrum1D(flux=flux_cube*u.Jy, wcs=cube_wcs,
                          uncertainty=uncertainties, mask=mask)

    # Initial model for fit.
    g1f = models.Gaussian1D(0.7*u.Jy, 4.65*u.um, 0.3*u.um, name='g1')
    g2f = models.Gaussian1D(2.0*u.Jy, 5.55*u.um, 0.3*u.um, name='g2')
    g3f = models.Gaussian1D(-2.*u.Jy, 8.15*u.um, 0.2*u.um, name='g3')
    zero_level = models.Const1D(1.*u.Jy, name='const1d')

    model_list = [g1f, g2f, g3f, zero_level]
    expression = "g1 + g2 + g3 + const1d"

    n_cpu = None
    # n_cpu = 1  # NOTE: UNCOMMENT TO DEBUG LOCALLY, AS NEEDED

    # Fit to all spaxels.
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", message="The fit may be unsuccessful.*")
        fitted_parameters, fitted_spectrum = fb.fit_model_to_spectrum(
            spectrum, model_list, expression, n_cpu=n_cpu)

    # Check that parameter results are formatted as expected.
    assert isinstance(fitted_parameters, list)
    assert len(fitted_parameters) == IMAGE_SIZE_X * IMAGE_SIZE_Y

    for m in fitted_parameters:
        if m['x'] == 3 and m['y'] == 2:
            fitted_model = m['model']

    assert isinstance(fitted_model[0].amplitude.value, np.float64)
    assert fitted_model[0].amplitude.unit == u.Jy

    assert isinstance(fitted_model[0], models.Gaussian1D)
    assert isinstance(fitted_model[0].mean.value, np.float64)
    assert fitted_model[0].mean.unit == u.um

    # Check that spectrum result is formatted as expected.
    assert isinstance(fitted_spectrum, Spectrum1D)
    assert len(fitted_spectrum.shape) == 3
    assert fitted_spectrum.shape == (IMAGE_SIZE_X, IMAGE_SIZE_Y, SPECTRUM_SIZE)

    # since were not loading data into app, results are just u.Jy not u.Jy / pix2
    assert fitted_spectrum.flux.unit == u.Jy
    assert not np.all(fitted_spectrum.flux.value == 0)

    # The important point here isn't to check the accuracy of the
    # fit, which was already tested elsewhere. We are mostly
    # interested here in checking the correctness of the data
    # packaging into the output products.

    assert_allclose(fitted_model[0].amplitude.value, 1.09, atol=TOL)
    assert_allclose(fitted_model[1].amplitude.value, 2.4, atol=TOL)
    assert_allclose(fitted_model[2].amplitude.value, -1.7, atol=TOL)

    assert_allclose(fitted_model[0].mean.value, 4.6, atol=TOL)
    assert_allclose(fitted_model[1].mean.value, 5.5, atol=TOL)
    assert_allclose(fitted_model[2].mean.value, 8.2, atol=TOL)

    assert_allclose(fitted_model[0].stddev.value, 0.2, atol=TOL)
    assert_allclose(fitted_model[1].stddev.value, 0.1, atol=TOL)
    assert_allclose(fitted_model[2].stddev.value, 0.1, atol=TOL)

    assert_allclose(fitted_model[3].amplitude.value, 4.0, atol=TOL)

    # Check that the fitted spectrum is masked correctly:
    assert_array_equal(fitted_spectrum.mask, mask)

    # Check I/O roundtrip.
    out_fn = tmp_path / "fitted_cube.fits"
    fitted_spectrum.write(out_fn, format="wcs1d-fits", hdu=1, flux_name="SCI", overwrite=True)
    flux_unit_str = fitted_spectrum.flux.unit.to_string(format="fits")
    coo_expected = fitted_spectrum.wcs.pixel_to_world(1, 0, 2)
    with fits.open(out_fn) as pf:
        assert len(pf) == 3
        assert pf[0].name == "PRIMARY"
        assert pf[1].name == "SCI"
        assert pf[1].header["BUNIT"] == flux_unit_str
        assert_allclose(pf[1].data, fitted_spectrum.flux.value)
        assert pf[2].name == "MASK"
        assert_array_equal(pf[2].data, mask)
        w = WCS(pf[1].header)
        coo = w.pixel_to_world(1, 0, 2)
        assert_allclose(coo[0].value, coo_expected[0].value)  # SpectralCoord
        assert_allclose([coo[1].ra.deg, coo[1].dec.deg],
                        [coo_expected[1].ra.deg, coo_expected[1].dec.deg])

    # Check Cubeviz roundtrip. This should automatically go through wcs1d-fits reader.
    cubeviz_helper.load_data(out_fn)
    assert len(cubeviz_helper.app.data_collection) == 3
    data_sci = cubeviz_helper.app.data_collection["fitted_cube.fits[SCI]"]
    flux_sci = data_sci.get_component("flux")
    assert_allclose(flux_sci.data, fitted_spectrum.flux.value)
    # now that the flux cube was loaded into cubeviz, there will be a factor
    # of pix2 applied to the flux unit
    assert flux_sci.units == f'{flux_unit_str} / pix2'
    coo = data_sci.coords.pixel_to_world(1, 0, 2)
    assert_allclose(coo[0].value, coo_expected[0].value)  # SpectralCoord
    assert_allclose([coo[1].ra.deg, coo[1].dec.deg],
                    [coo_expected[1].ra.deg, coo_expected[1].dec.deg])
    data_mask = cubeviz_helper.app.data_collection["fitted_cube.fits[MASK]"]
    flux_mask = data_mask.get_component("flux")
    assert_array_equal(flux_mask.data, mask)


def test_results_table(specviz_helper, spectrum1d):
    data_label = 'test'
    specviz_helper.load_data(spectrum1d, data_label=data_label)

    mf = specviz_helper.plugins['Model Fitting']
    mf.create_model_component('Linear1D')

    uc = specviz_helper.plugins['Unit Conversion']

    mf.add_results.label = 'linear model'
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', message='Model is linear in parameters.*')
        mf.calculate_fit(add_data=True)
    mf_table = mf.export_table()
    assert len(mf_table) == 1
    assert mf_table['equation'][-1] == 'L'
    assert mf_table.colnames == ['model', 'data_label', 'spectral_subset', 'equation',
                                 'L:slope_0', 'L:slope_0:unit',
                                 'L:slope_0:fixed', 'L:slope_0:std',
                                 'L:intercept_0', 'L:intercept_0:unit',
                                 'L:intercept_0:fixed', 'L:intercept_0:std']

    # verify units in table match the current display unit
    assert mf_table['L:intercept_0:unit'][-1] == uc.flux_unit

    mf.create_model_component('Gaussian1D')
    mf.add_results.label = 'composite model'
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', message='Model is linear in parameters.*')
        warnings.filterwarnings('ignore', message='The fit may be unsuccessful.*')
        mf.calculate_fit(add_data=True)
    mf_table = mf.export_table()
    assert len(mf_table) == 2
    assert mf_table['equation'][-1] == 'L+G'
    assert mf_table.colnames == ['model', 'data_label', 'spectral_subset', 'equation',
                                 'L:slope_0', 'L:slope_0:unit',
                                 'L:slope_0:fixed', 'L:slope_0:std',
                                 'L:intercept_0', 'L:intercept_0:unit',
                                 'L:intercept_0:fixed', 'L:intercept_0:std',
                                 'G:amplitude_1', 'G:amplitude_1:unit',
                                 'G:amplitude_1:fixed', 'G:amplitude_1:std',
                                 'G:mean_1', 'G:mean_1:unit',
                                 'G:mean_1:fixed', 'G:mean_1:std',
                                 'G:stddev_1', 'G:stddev_1:unit',
                                 'G:stddev_1:fixed', 'G:stddev_1:std']

    mf.remove_model_component('G')
    assert len(mf_table) == 2

    # verify Spectrum1D model fitting plugin and table can handle spectral density conversions
    uc.flux_unit = 'erg / (Angstrom s cm2)'
    mf.reestimate_model_parameters()

    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', message='Model is linear in parameters.*')
        mf.calculate_fit(add_data=True)

    assert mf_table['L:intercept_0:unit'][-1] == uc.flux_unit


def test_equation_validation(specviz_helper, spectrum1d):
    data_label = 'test'
    specviz_helper.load_data(spectrum1d, data_label=data_label)

    mf = specviz_helper.plugins['Model Fitting']
    mf.create_model_component('Const1D')
    mf.create_model_component('Linear1D')

    assert mf.equation == 'C+L'
    assert mf._obj.model_equation_invalid_msg == ''

    mf.equation = 'L+'
    assert mf._obj.model_equation_invalid_msg == 'incomplete equation.'

    mf.equation = 'L+C'
    assert mf._obj.model_equation_invalid_msg == ''

    mf.equation = 'L+CC'
    assert mf._obj.model_equation_invalid_msg == 'CC is not an existing model component.'

    mf.equation = 'L+CC+DD'
    assert mf._obj.model_equation_invalid_msg == 'CC, DD are not existing model components.'

    mf.equation = ''
    assert mf._obj.model_equation_invalid_msg == 'model equation is required.'


def test_incompatible_units(specviz_helper, spectrum1d):
    data_label = 'test'
    specviz_helper.load_data(spectrum1d, data_label=data_label)

    mf = specviz_helper.plugins['Model Fitting']
    mf.create_model_component('Linear1D')

    mf.add_results.label = 'model native units'
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', message='Model is linear in parameters.*')
        mf.calculate_fit(add_data=True)

    uc = specviz_helper.plugins['Unit Conversion']
    assert uc.spectral_unit.selected == "Angstrom"
    uc.spectral_unit = u.Hz

    assert 'L is currently disabled' in mf._obj.model_equation_invalid_msg
    mf.add_results.label = 'frequency units'
    with pytest.raises(ValueError, match="model equation is invalid.*"):
        mf.calculate_fit(add_data=True)

    mf.reestimate_model_parameters()
    assert mf._obj.model_equation_invalid_msg == ''
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', message='Model is linear in parameters.*')
        mf.calculate_fit(add_data=True)


def test_cube_fit_with_nans(cubeviz_helper):
    flux = np.ones((7, 8, 9)) * u.nJy
    flux[:, :, 0] = np.nan
    spec = Spectrum1D(flux=flux)
    cubeviz_helper.load_data(spec, data_label="test")

    mf = cubeviz_helper.plugins["Model Fitting"]
    mf.cube_fit = True
    mf.create_model_component("Const1D")
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', message='Model is linear in parameters.*')
        mf.calculate_fit()
    result = cubeviz_helper.app.data_collection['model']
    assert np.all(result.get_component("flux").data == 1)

    # Switch back to non-cube fit, check that units are marked incompatible
    mf.cube_fit = False
    assert mf._obj.component_models[0]['compat_display_units'] is False


def test_cube_fit_with_subset_and_nans(cubeviz_helper):
    # Also test with existing mask
    flux = np.ones((7, 8, 9)) * u.nJy
    flux[:, :, 0] = np.nan
    spec = Spectrum1D(flux=flux)
    spec.flux[5, 5, 7] = 10 * u.nJy
    cubeviz_helper.load_data(spec, data_label="test")

    sv = cubeviz_helper.app.get_viewer('spectrum-viewer')
    sv.apply_roi(XRangeROI(0, 5))

    mf = cubeviz_helper.plugins["Model Fitting"]
    mf.cube_fit = True
    mf.spectral_subset = 'Subset 1'
    mf.create_model_component("Const1D")
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', message='Model is linear in parameters.*')
        mf.calculate_fit()
    result = cubeviz_helper.app.data_collection['model']
    assert np.all(result.get_component("flux").data == 1)


def test_fit_with_count_units(cubeviz_helper):
    flux = np.random.random((7, 8, 9)) * u.count
    spectral_axis = np.linspace(4000, 5000, flux.shape[-1]) * u.AA

    spec = Spectrum1D(flux=flux, spectral_axis=spectral_axis)
    cubeviz_helper.load_data(spec, data_label="test")

    mf = cubeviz_helper.plugins["Model Fitting"]
    mf.cube_fit = True
    mf.create_model_component("Const1D")

    # ensures specutils.Spectrum1D.with_flux_unit has access to Jdaviz custom equivalencies for
    # PIX^2 unit
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', message='Model is linear in parameters.*')
        mf.calculate_fit()

    assert mf._obj.component_models[0]['parameters'][0]['unit'] == 'ct / pix2'

    model_flux = cubeviz_helper.app.data_collection[-1].get_component('flux')
    assert model_flux.units == 'ct / pix2'


@pytest.mark.parametrize("solid_angle_unit", [u.sr, PIX2])
def test_cube_fit_after_unit_change(cubeviz_helper, solid_angle_unit):
    cube = _create_spectrum1d_cube_with_fluxunit(fluxunit=u.Jy / solid_angle_unit, shape=(10, 4, 5),
                                                 with_uncerts=True)
    cubeviz_helper.load_data(cube, data_label="test")
    solid_angle_string = str(solid_angle_unit)

    uc = cubeviz_helper.plugins['Unit Conversion']
    mf = cubeviz_helper.plugins['Model Fitting']
    uc.flux_unit = "MJy"
    mf.cube_fit = True

    mf.create_model_component("Const1D")
    # Check that the parameter is using the current units when initialized
    assert mf._obj.component_models[0]['parameters'][0]['unit'] == f'MJy / {solid_angle_string}'

    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', message='Model is linear in parameters.*')
        mf.calculate_fit()

    expected_result_slice = np.array([[9.00e-05, 9.50e-05, 1.00e-04, 1.05e-04],
                                      [9.10e-05, 9.60e-05, 1.01e-04, 1.06e-04],
                                      [9.20e-05, 9.70e-05, 1.02e-04, 1.07e-04],
                                      [9.30e-05, 9.80e-05, 1.03e-04, 1.08e-04],
                                      [9.40e-05, 9.90e-05, 1.04e-04, 1.09e-04]])

    model_flux = cubeviz_helper.app.data_collection[-1].get_component('flux')
    assert model_flux.units == f'MJy / {solid_angle_string}'
    assert np.allclose(model_flux.data[:, :, 1], expected_result_slice)

    # Switch back to Jy, see that the component didn't change but the output does
    uc.flux_unit = 'Jy'
    assert mf._obj.component_models[0]['parameters'][0]['unit'] == f'MJy / {solid_angle_string}'
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', message='Model is linear in parameters.*')
        mf.calculate_fit()

    model_flux = cubeviz_helper.app.data_collection[-1].get_component('flux')
    assert model_flux.units == f'Jy / {solid_angle_string}'
    assert np.allclose(model_flux.data[:, :, 1], expected_result_slice * 1e6)

    # ensure conversions that require the spectral axis/translations are handled by the plugin
    uc.spectral_y_type = 'Surface Brightness'
    uc.flux_unit = 'erg / (Angstrom s cm2)'

    mf.reestimate_model_parameters()

    if solid_angle_string == 'sr':
        expected_unit_string = f'erg / (Angstrom s {solid_angle_string} cm2)'
    else:
        expected_unit_string = f'erg / (Angstrom s cm2 {solid_angle_string})'

    assert mf._obj.component_models[0]['parameters'][0]['unit'] == expected_unit_string

    # running this ensures specutils.Spectrum1D.with_flux_unit has Jdaviz custom equivalencies
    # for spectral axis conversions and scale factor translations
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', message='Model is linear in parameters.*')
        mf.calculate_fit()

    model_flux = cubeviz_helper.app.data_collection[-1].get_component('flux')
    assert model_flux.units == expected_unit_string



import asdf
import numpy as np
import astropy.units as u

from jdaviz.configs.imviz.tests.utils import create_example_gwcs


def test_asdf_not_rdm(imviz_helper):
    # test support for ASDF files that look like Roman files
    # for users with or without roman_datamodels:
    in_unit = u.Jy
    in_data = np.arange(16, dtype=np.float32).reshape((4, 4)) * in_unit
    tree = {
        'roman': {
            'data': in_data,
            'meta': {
                'wcs': create_example_gwcs((4, 4))
            },
        },
    }

    af = asdf.AsdfFile(tree=tree)
    imviz_helper.load_data(af)
    out_component = imviz_helper.app.data_collection[0].get_component('DATA')
    np.testing.assert_array_equal(in_data.value, out_component.data)
    assert str(in_unit) == out_component.units



import os

import asdf
import numpy as np
from astropy import units as u
from astropy.io import fits
from astropy.nddata import NDData
from astropy.wcs import WCS
from astropy.utils.data import cache_contents

from glue.core.data import Component, Data
from gwcs.wcs import WCS as GWCS
from stdatamodels import asdf_in_fits

from jdaviz.core.registries import data_parser_registry
from jdaviz.core.events import SnackbarMessage
from jdaviz.utils import (
    standardize_metadata, standardize_roman_metadata,
    PRIHDR_KEY, _wcs_only_label, download_uri_to_path
)

try:
    from roman_datamodels import datamodels as rdd
except ImportError:
    HAS_ROMAN_DATAMODELS = False
else:
    HAS_ROMAN_DATAMODELS = True

__all__ = ['parse_data']

INFO_MSG = ("The file contains more viewable extensions. Add the '[*]' suffix"
            " to the file name to load all of them.")


def prep_data_layer_as_dq(data):
    # nans are used to mark "good" flags in the DQ colormap, so
    # convert DQ array to float to support nans:
    for component_id in data.main_components:
        if component_id.label.startswith("DQ"):
            break

    cid = data.get_component(component_id)
    data_arr = np.float32(cid.data)
    data_arr[data_arr == 0] = np.nan
    data.update_components({cid: data_arr})


@data_parser_registry("imviz-data-parser")
def parse_data(app, file_obj, ext=None, data_label=None,
               parent=None, cache=None, local_path=None, timeout=None):
    """Parse a data file into Imviz.

    Parameters
    ----------
    app : `~jdaviz.app.Application`
        The application-level object used to reference the viewers.

    file_obj : str or obj
        The path to an image data file or FITS HDUList or image object.

    ext : str, tuple, or `None`, optional
        FITS extension, if given. Examples: ``'SCI'`` or ``('SCI', 1)``

    data_label : str, optional
        The label to be applied to the Glue data component.

    parent : str, optional
        Data label for "parent" data to associate with the loaded data as "child".

    cache : None, bool, or str
        Cache the downloaded file if the data are retrieved by a query
        to a URL or URI.

    local_path : str, optional
        Cache remote files to this path. This is only used if data is
        requested from `astroquery.mast`.

    timeout : float, optional
        If downloading from a remote URI, set the timeout limit for
        remote requests in seconds (passed to
        `~astropy.utils.data.download_file` or
        `~astroquery.mast.Conf.timeout`).
    """
    if isinstance(file_obj, str):
        if data_label is None:
            data_label = os.path.splitext(os.path.basename(file_obj))[0]

        # try parsing file_obj as a URI/URL:
        file_obj = download_uri_to_path(
            file_obj, cache=cache, local_path=local_path, timeout=timeout
        )

        # If file_obj is a path to a cached file from
        # astropy.utils.data.download_file, the path has no file extension.
        # Here we check if the file is in the download cache, and if it is,
        # we look up the file extension from the source URL:
        if file_obj.endswith('contents'):
            path_to_url_mapping = {v: k for k, v in cache_contents().items() if file_obj}
            source_url = path_to_url_mapping[file_obj]
            # file_obj_lower is only used for checking extensions,
            # file_obj is passed for parsing and is not modified here:
            file_obj_lower = source_url.split('/')[-1].lower()
        else:
            file_obj_lower = file_obj.lower()

        if file_obj_lower.endswith(('.jpg', '.jpeg', '.png')):
            from skimage.io import imread
            from skimage.color import rgb2gray, rgba2rgb
            im = imread(file_obj)
            if im.shape[2] == 4:
                pf = rgb2gray(rgba2rgb(im))
            else:  # Assume RGB
                pf = rgb2gray(im)
            pf = pf[::-1, :]  # Flip it
            _parse_image(app, pf, data_label, ext=ext, parent=parent)

        elif file_obj_lower.endswith('.asdf'):
            try:
                if HAS_ROMAN_DATAMODELS:
                    with rdd.open(file_obj) as pf:
                        _parse_image(app, pf, data_label, ext=ext, parent=parent)
            except TypeError:
                # if roman_datamodels cannot parse the file, load it with asdf:
                with asdf.open(file_obj) as af:
                    _parse_image(app, af, data_label, ext=ext, parent=parent)

        elif file_obj_lower.endswith('.reg'):
            # This will load DS9 regions as Subset but only if there is already data.
            app.get_tray_item_from_name('g-subset-tools').import_region(file_obj,
                                                                        combination_mode='new')

        else:  # Assume FITS
            with fits.open(file_obj) as pf:
                available_extensions = [hdu.name for hdu in pf]

                # if FITS file contains SCI and DQ extensions, assume the
                # parent for the DQ is SCI:
                if (
                    'SCI' in available_extensions and
                    ext == 'DQ' and parent is None
                ):
                    loaded_data_labels = [data.label for data in app.data_collection]
                    latest_sci_extension = [label for label in loaded_data_labels
                                            if label.endswith('[DATA]')][-1]
                    parent = latest_sci_extension

                _parse_image(app, pf, data_label, ext=ext, parent=parent)
    else:
        _parse_image(app, file_obj, data_label, ext=ext, parent=parent)


def get_image_data_iterator(app, file_obj, data_label, ext=None):
    """This function is for internal use, so other viz can also extract image data
    like Imviz does.
    """
    if isinstance(file_obj, fits.HDUList):
        if 'ASDF' in file_obj:  # JWST ASDF-in-FITS
            # Load multiple extensions
            if ext == '*' or isinstance(ext, (tuple, list)):
                data_iter = _jwst_all_to_glue_data(file_obj, data_label, load_extensions=ext)

            # Load only specified extension
            else:
                data_iter = _jwst_to_glue_data(file_obj, ext, data_label)

                # if more than one viewable extension is found in the file,
                # issue info message.
                _info_nextensions(app, file_obj)

        elif ext == '*' or isinstance(ext, (tuple, list)):  # Load multiple extensions
            data_iter = _hdus_to_glue_data(file_obj, data_label, ext=ext)

        elif ext is not None:  # Load just the EXT user wants
            hdu = file_obj[ext]
            _validate_fits_image2d(hdu)
            data_iter = _hdu_to_glue_data(hdu, data_label, hdulist=file_obj)

        else:  # Load first image extension found
            found = False
            for hdu in file_obj:
                if _validate_fits_image2d(hdu, raise_error=False):
                    data_iter = _hdu_to_glue_data(hdu, data_label, hdulist=file_obj)
                    found = True

                    # if more than one viewable extension is found in the file,
                    # issue info message.
                    _info_nextensions(app, file_obj)

                    break

            if not found:
                raise ValueError(f'{file_obj} does not have any FITS image')

    elif isinstance(file_obj, (fits.ImageHDU, fits.CompImageHDU, fits.PrimaryHDU)):
        # NOTE: ext is not used here. It only means something if HDUList is given.
        _validate_fits_image2d(file_obj)
        data_iter = _hdu_to_glue_data(file_obj, data_label)

    elif isinstance(file_obj, NDData):
        if file_obj.meta.get(_wcs_only_label, False):
            data_iter = _wcsonly_to_glue_data(file_obj, data_label)
        else:
            data_iter = _nddata_to_glue_data(file_obj, data_label)

    elif isinstance(file_obj, np.ndarray):
        data_iter = _ndarray_to_glue_data(file_obj, data_label)

    # load Roman 2D datamodels:
    elif HAS_ROMAN_DATAMODELS and isinstance(file_obj, rdd.DataModel):
        data_iter = _roman_2d_to_glue_data(file_obj, data_label, ext=ext)

    # load ASDF files that may not validate as Roman datamodels:
    elif isinstance(file_obj, asdf.AsdfFile):
        data_iter = _roman_asdf_2d_to_glue_data(file_obj, data_label, ext=ext)

    else:
        raise NotImplementedError(f'Imviz does not support {file_obj}')

    return data_iter


def _parse_image(app, file_obj, data_label, ext=None, parent=None):
    if app is None:
        raise ValueError("app is None, cannot proceed")
    if data_label is None:
        data_label = app.return_data_label(file_obj, ext, alt_name="image_data")
    data_iter = get_image_data_iterator(app, file_obj, data_label, ext=ext)

    # Save the SCI extension to this list:
    sci_ext = None

    for data, data_label in data_iter:

        # if the science extension hasn't been identified yet, do so here:
        if sci_ext is None and data.ndim == 2 and ('[DATA' in data_label or '[SCI' in data_label):
            sci_ext = data_label

        if isinstance(data.coords, GWCS) and (data.coords.bounding_box is not None):
            # keep a copy of the original bounding box so we can detect
            # when extrapolating beyond, but then remove the bounding box
            # so that image layers are not cropped.
            # NOTE: if extending this beyond GWCS, the mouseover logic
            # for outside_*_bounding_box should also be updated.
            data.coords._orig_bounding_box = data.coords.bounding_box
            data.coords.bounding_box = None
        if not data.meta.get(_wcs_only_label, False):
            data_label = app.return_data_label(data_label, alt_name="image_data")

        # TODO: generalize/centralize this for use in other configs too
        if '[DQ' in data_label:
            prep_data_layer_as_dq(data)

        if parent is not None:
            parent_data_label = parent
        elif '[DQ' in data_label:
            parent_data_label = sci_ext
        else:
            parent_data_label = None

        app.add_data(data, data_label, parent=parent_data_label)

    # Do not link image data here. We do it at the end in Imviz.load_data()


def _info_nextensions(app, file_obj):
    if _count_image2d_extensions(file_obj) > 1:
        info_msg = SnackbarMessage(INFO_MSG, color="info", timeout=8000, sender=app)
        app.hub.broadcast(info_msg)


def _count_image2d_extensions(file_obj):
    count = 0
    for hdu in file_obj:
        if _validate_fits_image2d(hdu, raise_error=False):
            count += 1
    return count


def _validate_fits_image2d(hdu, raise_error=True):
    valid = hdu.data is not None and hdu.is_image and hdu.data.ndim == 2
    if not valid and raise_error:
        if hdu.data is not None and hdu.is_image:
            ndim = hdu.data.ndim
        else:
            ndim = None
        raise ValueError(
            f'Imviz cannot load this HDU ({hdu}): '
            f'has_data={hdu.data is not None}, is_image={hdu.is_image}, '
            f'name={hdu.name}, ver={hdu.ver}, ndim={ndim}')
    return valid


def _validate_bunit(bunit, raise_error=True):
    # TODO: Handle weird FITS BUNIT values here, as needed.
    if bunit == 'ELECTRONS/S':
        valid = 'electron/s'
    elif bunit == 'ELECTRONS':
        valid = 'electron'
    else:
        try:
            u.Unit(bunit)
        except Exception:
            if raise_error:
                raise
            valid = ''
        else:
            valid = bunit
    return valid


# ---- Functions that handle input from JWST FITS files -----

def _jwst_all_to_glue_data(file_obj, data_label, load_extensions='*'):
    for hdu in file_obj:
        if (
            _validate_fits_image2d(hdu, raise_error=False) and
            (load_extensions == '*' or hdu.name in load_extensions)
        ):

            ext = hdu.name.lower()
            if ext == 'sci':
                ext = 'data'

            data, new_data_label = _jwst2data(file_obj, ext, data_label)

            yield data, new_data_label


def _jwst_to_glue_data(file_obj, ext, data_label):
    # Translate FITS extension into JWST ASDF convention.
    if ext is None:
        ext = 'data'
        fits_ext = 'sci'
    else:
        if isinstance(ext, tuple):
            ext = ext[0]  # EXTVER means nothing in ASDF
        ext = ext.lower()
        if ext in ('sci', 'asdf'):
            ext = 'data'
            fits_ext = 'sci'
        else:
            fits_ext = ext

    _validate_fits_image2d(file_obj[fits_ext])
    data, new_data_label = _jwst2data(file_obj, ext, data_label)

    yield data, new_data_label


def _jwst2data(file_obj, ext, data_label):
    comp_label = ext.upper()
    new_data_label = f'{data_label}[{comp_label}]'
    data = Data(label=new_data_label)
    unit_attr = f'bunit_{ext}'

    try:
        # This is very specific to JWST pipeline image output.
        with asdf_in_fits.open(file_obj) as af:
            dm = af.tree
            dm_meta = af.tree["meta"]
            data.meta.update(standardize_metadata(dm_meta))

            if unit_attr in dm_meta:
                bunit = _validate_bunit(dm_meta[unit_attr], raise_error=False)
            else:
                bunit = ''

            # This is instance of gwcs.WCS, not astropy.wcs.WCS
            if 'wcs' in dm_meta:
                data.coords = dm_meta['wcs']

            imdata = dm[ext]
            component = Component.autotyped(imdata, units=bunit)

            # Might have bad GWCS. If so, we exclude it.
            try:
                data.add_component(component=component, label=comp_label)
            except Exception:  # pragma: no cover
                data.coords = None
                data.add_component(component=component, label=comp_label)

    # TODO: Do not need this when jwst.datamodels finally its own package.
    # This might happen for grism image; fall back to FITS loader without WCS.
    except Exception:
        if ext == 'data':
            ext = 'sci'
        hdu = file_obj[ext]
        return _hdu2data(hdu, data_label, file_obj, include_wcs=False)

    return data, new_data_label


# ---- Functions that handle input from Roman ASDF files -----

def _roman_2d_to_glue_data(file_obj, data_label, ext=None):

    if ext == '*':
        # NOTE: Update as needed. Should cover all the image extensions available.
        ext_list = ('data', 'dq', 'err', 'var_poisson', 'var_rnoise')
    elif ext is None:
        ext_list = ('data', )
    elif isinstance(ext, (list, tuple)):
        ext_list = ext
    else:
        ext_list = (ext, )

    meta = standardize_roman_metadata(file_obj)
    coords = getattr(getattr(file_obj, 'meta', {}), 'wcs', None)

    for cur_ext in ext_list:
        comp_label = cur_ext.upper()
        new_data_label = f'{data_label}[{comp_label}]'
        data = Data(coords=coords, label=new_data_label)

        # This could be a quantity or a ndarray:
        ext_values = getattr(file_obj, cur_ext)
        bunit = getattr(ext_values, 'unit', '')
        component = Component.autotyped(np.array(ext_values), units=bunit)
        data.add_component(component=component, label=comp_label)
        data.meta.update(meta)

        if comp_label == 'dq':
            prep_data_layer_as_dq(data)

        yield data, new_data_label


def _roman_asdf_2d_to_glue_data(file_obj, data_label, ext=None):
    if ext == '*':
        # NOTE: Update as needed. Should cover all the image extensions available.
        ext_list = ('data', 'dq', 'err', 'var_poisson', 'var_rnoise')
    elif ext is None:
        ext_list = ('data', )
    elif isinstance(ext, (list, tuple)):
        ext_list = ext
    else:
        ext_list = (ext, )

    roman = file_obj.tree.get('roman')
    meta = roman.get('meta', {})
    coords = meta.get('wcs', None)

    for cur_ext in ext_list:
        if cur_ext in roman:
            comp_label = cur_ext.upper()
            new_data_label = f'{data_label}[{comp_label}]'
            data = Data(coords=coords, label=new_data_label)

            # This could be a quantity or a ndarray:
            ext_values = roman.get(cur_ext)
            bunit = getattr(ext_values, 'unit', '')
            component = Component(np.array(ext_values), units=bunit)
            data.add_component(component=component, label=comp_label)
            data.meta.update(standardize_metadata(dict(meta)))
            if comp_label == 'DQ':
                prep_data_layer_as_dq(data)

            yield data, new_data_label


# ---- Functions that handle input from non-JWST, non-Roman FITS files -----

def _hdu_to_glue_data(hdu, data_label, hdulist=None):
    data, data_label = _hdu2data(hdu, data_label, hdulist)
    yield data, data_label


def _hdus_to_glue_data(file_obj, data_label, ext=None):
    for hdu in file_obj:
        if ext is None or ext == '*' or hdu.name in ext:
            if _validate_fits_image2d(hdu, raise_error=False):
                data, new_data_label = _hdu2data(hdu, data_label, file_obj)
                yield data, new_data_label


def _hdu2data(hdu, data_label, hdulist, include_wcs=True):
    if 'BUNIT' in hdu.header:
        bunit = _validate_bunit(hdu.header['BUNIT'], raise_error=False)
    else:
        bunit = ''

    comp_label = f'{hdu.name.upper()},{hdu.ver}'
    new_data_label = f'{data_label}[{comp_label}]'

    data = Data(label=new_data_label)
    if hdulist is not None and hdu.name != 'PRIMARY' and 'PRIMARY' in hdulist:
        data.meta[PRIHDR_KEY] = standardize_metadata(hdulist['PRIMARY'].header)
    data.meta.update(standardize_metadata(hdu.header))
    if include_wcs:
        data.coords = WCS(hdu.header, hdulist)
    component = Component.autotyped(hdu.data, units=bunit)
    data.add_component(component=component, label=comp_label)

    return data, new_data_label


# ---- Functions that handle input from arrays -----

def _nddata_to_glue_data(ndd, data_label):
    if ndd.data.ndim != 2:
        raise ValueError(f'Imviz cannot load this NDData with ndim={ndd.data.ndim}')

    for attrib in ('data', 'mask', 'uncertainty'):
        arr = getattr(ndd, attrib)
        if arr is None:
            continue
        comp_label = attrib.upper()
        cur_label = f'{data_label}[{comp_label}]'
        cur_data = Data(label=cur_label)
        cur_data.meta.update(standardize_metadata(ndd.meta))
        if ndd.wcs is not None:
            cur_data.coords = ndd.wcs
        raw_arr = arr
        if attrib == 'data':
            bunit = ndd.unit or ''
        elif attrib == 'uncertainty':
            raw_arr = arr.array
            bunit = arr.unit or ''
        else:
            bunit = ''
        component = Component.autotyped(raw_arr, units=bunit)
        cur_data.add_component(component=component, label=comp_label)
        yield cur_data, cur_label


def _ndarray_to_glue_data(arr, data_label):
    if arr.ndim != 2:
        raise ValueError(f'Imviz cannot load this array with ndim={arr.ndim}')

    data = Data(label=data_label)
    component = Component.autotyped(arr)
    data.add_component(component=component, label='DATA')
    yield (data, data_label)


# ---- Functions that handle WCS-only data -----

def _wcsonly_to_glue_data(ndd, data_label):
    """Return Data given NDData containing WCS-only data."""
    arr = ndd.data
    data = Data(label=data_label)
    data.meta.update(standardize_metadata(ndd.meta))
    data.coords = ndd.wcs
    component = Component.autotyped(arr, units="")
    data.add_component(component=component, label="DATA")
    yield (data, data_label)



# Tests automatic config detection against our example notebook data

from pathlib import Path

import pytest
from astroquery.mast import Observations
from astropy.utils.data import download_file

from jdaviz import open as jdaviz_open
from jdaviz.cli import ALL_JDAVIZ_CONFIGS
from jdaviz.configs import Specviz2d, Cubeviz, Imviz, Specviz
from jdaviz.core.launcher import Launcher, STATUS_HINTS


AUTOCONFIG_EXAMPLES = (
    ("mast:JWST/product/jw02732-o004_t004_miri_ch1-shortmediumlong_x1d.fits", Specviz),
    ("mast:JWST/product/jw01538160001_16101_00004_nrs1_s2d.fits", Specviz2d),
    ("mast:JWST/product/jw02727-o002_t062_nircam_clear-f090w_i2d.fits", Imviz),
    ("mast:JWST/product/jw02732-o004_t004_miri_ch1-shortmediumlong_s3d.fits", Cubeviz),
    ("https://stsci.box.com/shared/static/28a88k1qfipo4yxc4p4d40v4axtlal8y.fits", Cubeviz)
    # Check that MaNGA cubes go to cubeviz. This file is originally from:
    # https://data.sdss.org/sas/dr14/manga/spectro/redux/v2_1_2/7495/stack/manga-7495-12704-LOGCUBE.fits.gz)
)


@pytest.mark.remote_data
@pytest.mark.slow
@pytest.mark.filterwarnings('ignore')
@pytest.mark.parametrize('uris', AUTOCONFIG_EXAMPLES)
def test_autoconfig(uris, tmp_path):
    uri = uris[0]
    helper_class = uris[1]

    local_path = str(tmp_path / Path(uri).name)

    kwargs = dict(cache=True, show=False)
    if uri.startswith('mast'):
        kwargs['local_path'] = local_path

    viz_helper = jdaviz_open(uri, **kwargs)
    assert isinstance(viz_helper, helper_class)
    assert len(viz_helper.app.data_collection) > 0


@pytest.mark.remote_data
@pytest.mark.filterwarnings('ignore')
def test_launcher(tmp_path):
    launcher = Launcher(None, ALL_JDAVIZ_CONFIGS)

    # Test starting state
    assert launcher.hint == STATUS_HINTS['idle']
    assert launcher.compatible_configs == ALL_JDAVIZ_CONFIGS

    # Test invalid filepath
    launcher.filepath = "ThisIsAFakeFileNameThatShouldNotExist"
    assert launcher.hint == STATUS_HINTS['invalid path']
    assert len(launcher.compatible_configs) == 0

    # Test reseting to empty state
    launcher.filepath = ""
    assert launcher.hint == STATUS_HINTS['idle']
    assert launcher.compatible_configs == ALL_JDAVIZ_CONFIGS

    # Test with real files
    for uri, config in AUTOCONFIG_EXAMPLES:
        if uri.startswith("mast:"):
            download_path = str(tmp_path / Path(uri).name)
            Observations.download_file(uri, local_path=download_path)
        elif uri.startswith("http"):
            download_path = download_file(uri, cache=True, timeout=100)
        launcher.filepath = download_path
        # In testing, setting the filepath will stall until identifying is complete
        # No need to be concerned for race condition here
        assert launcher.hint == STATUS_HINTS['id ok']
        assert launcher.compatible_configs == [config().__class__.__name__.lower()]

    # Test reseting to empty state
    launcher.filepath = ""
    assert launcher.hint == STATUS_HINTS['idle']
    assert launcher.compatible_configs == ALL_JDAVIZ_CONFIGS



import os
import traitlets

from glue.config import viewer_tool
from glue.core import HubListener
from glue.icons import icon_path
from glue.viewers.common.tool import CheckableTool
from glue_jupyter.common.toolbar_vuetify import BasicJupyterToolbar, read_icon

from jdaviz.core.events import (AddDataMessage, RemoveDataMessage,
                                ViewerAddedMessage, ViewerRemovedMessage,
                                SpectralMarksChangedMessage, CatalogResultsChangedMessage)

__all__ = ['NestedJupyterToolbar']


class NestedJupyterToolbar(BasicJupyterToolbar, HubListener):
    template_file = (__file__, 'toolbar_nested.vue')

    # defined in BasicJupyterToolbar:
    # active_tool = traitlets.Instance
    # tools_data = traitlets.Dict
    # active_tool_id = traitlets.Any

    # whether to show a popup menu
    show_suboptions = traitlets.Bool(False).tag(sync=True)
    close_on_click = traitlets.Bool(False).tag(sync=True)
    # which submenu to show when show_suboptions is True
    suboptions_ind = traitlets.Int(0).tag(sync=True)
    # absolute positions to display the menu
    suboptions_x = traitlets.Float().tag(sync=True)
    suboptions_y = traitlets.Float().tag(sync=True)

    def __init__(self, viewer, tools_nested, default_tool_priority=[]):
        super().__init__(viewer)
        self.viewer = viewer
        self._max_menu_ind = len(tools_nested)

        # iterate through the nested list.  The first item in each entry
        # is treated as the default primary tool of that subcategory,
        # with each additional entry available from a dropdown.  For
        # backwards compatibility with BasicJupyterToolbar, single strings
        # will be treated as having no submenu.
        for menu_ind, subtools in enumerate(tools_nested):
            if isinstance(subtools, str):
                subtools = [subtools]
            for i, tool_id in enumerate(subtools):
                tool_cls = viewer_tool.members[tool_id]
                tool = tool_cls(viewer)
                self.add_tool(tool,
                              menu_ind=menu_ind,
                              has_suboptions=len(subtools) > 1,
                              primary=i == 0,
                              visible=True)

        # default_tool_priority allows falling back on an existing tool
        # if its the primary tool.  If no items in default_tool_priority
        # are currently "primary", then either no tool will be selected
        # or will fallback on BasicJupyterToolbar's handling of
        # viewer._default_mouse_mode_cls (which will not show that tool as active).
        self.default_tool_priority = default_tool_priority

        # handle logic for tool visibilities (which will also handle setting the primary
        # to something other than the first entry, if necessary)
        # NOTE: this will also call _handle_default_tool
        self._update_tool_visibilities()

        # toolbars in the main app viewers need to respond to the data-collection, etc,
        # but those in plugins do not
        if hasattr(self.viewer, 'hub'):
            for msg in (AddDataMessage, RemoveDataMessage, ViewerAddedMessage, ViewerRemovedMessage,
                        SpectralMarksChangedMessage, CatalogResultsChangedMessage):
                self.viewer.hub.subscribe(self, msg,
                                          handler=lambda _: self._update_tool_visibilities())

    def _is_visible(self, tool_id):
        # tools can optionally implement self.is_visible(). If not NotImplementedError
        # the tool will always be visible
        if hasattr(self.tools[tool_id], 'is_visible'):
            return self.tools[tool_id].is_visible()
        return True

    def _update_tool_visibilities(self):
        needs_deactivate_active = False
        for menu_ind in range(self._max_menu_ind):
            has_primary = False
            n_visible = 0
            if self.active_tool_id:
                current_primary_active = self.tools_data[self.active_tool_id]['menu_ind'] == menu_ind  # noqa
            else:
                current_primary_active = False
            for tool_id, info in self.tools_data.items():
                if info['menu_ind'] != menu_ind:
                    continue
                visible = self._is_visible(tool_id)
                if visible:
                    n_visible += 1

                if tool_id == self.active_tool_id:
                    # then the primary was already set by which tool is active
                    if visible:
                        # then keep this as primary
                        primary = True
                    else:
                        # then the currently active tool is being removed, so we need to deactivate
                        # the tool and allow the default logic to be triggered
                        primary = False
                        needs_deactivate_active = True
                elif current_primary_active and self._is_visible(self.active_tool_id):
                    # then we are keeping the previous primary
                    primary = False
                else:
                    # then there is no primary already in this submenu, so the first visible
                    # entry will be selected as primary
                    primary = visible and not has_primary

                if primary:
                    has_primary = True

                self.tools_data[tool_id] = {**info,
                                            'primary': primary,
                                            'visible': visible}

            # update has_suboptions for all entries in this menu
            for tool_id, info in self.tools_data.items():
                if info['menu_ind'] != menu_ind:
                    continue
                self.tools_data[tool_id] = {**self.tools_data[tool_id],
                                            'has_suboptions': n_visible > 1}

        # mutation to dictionary needs to be manually sent to update the UI
        self.send_state("tools_data")
        if needs_deactivate_active:
            self.active_tool_id = None
        self._handle_default_tool()

    def _handle_default_tool(self):
        # default to the first item in the default_tool_priority list that is currently
        # already primary
        for tool_id in self.default_tool_priority:
            if tool_id not in self.tools_data:
                continue
            if self.tools_data[tool_id]['primary'] and self.tools_data[tool_id]['visible']:
                self.active_tool_id = tool_id
                break

    @traitlets.observe('active_tool_id')
    def _on_change_v_model(self, event):
        super()._on_change_v_model(event)

        if event['new'] is None and event['old'] not in self.default_tool_priority:
            # then we're unchecking a non-default tool
            self._handle_default_tool()
        elif event['new'] is not None and not isinstance(self.tools[event['new']], CheckableTool):
            # then we're clicking on a non-checkable tool and want to default to the previous
            if event['old'] is not None:
                self.active_tool_id = event['old']

    def _select_tool(self, tool_id, menu_ind):
        for search_tool_id, info in self.tools_data.items():
            if info['menu_ind'] == menu_ind and info['primary']:
                prev_id = search_tool_id
                prev_info = info
                break
        else:
            raise ValueError("could not find previous selection")

        if isinstance(self.tools[tool_id], CheckableTool):
            # only switch to primary if its actually checkable, otherwise
            # just activate once
            self.tools_data = {
                **self.tools_data,
                prev_id: {
                    **prev_info,
                    'primary': False
                },
                tool_id: {
                    **self.tools_data[tool_id],
                    'primary': True
                }
            }

        # and finally, set to be the active tool (this triggers _on_change_v_model which in turn
        # triggers BasicJupyterToolbar._on_change_active_tool)
        self.active_tool_id = tool_id

    def select_tool(self, tool_id):
        # find the previous primary tool in the same menu_ind
        menu_ind = self.tools_data[tool_id]['menu_ind']
        self._select_tool(tool_id, menu_ind)

    def add_tool(self, tool, menu_ind, has_suboptions=True, primary=False, visible=True):
        # NOTE: this method is essentially copied from glue-jupyter's BasicJupyterToolbar,
        # but we need extra values in the tools_data dictionary.  We could call super(),
        # but then that would create tools_data twice, which would then cause
        # issues/re-rerendering
        self.tools[tool.tool_id] = tool
        if os.path.exists(tool.icon):
            path = tool.icon
        else:
            path = icon_path(tool.icon, icon_format='svg')
        self.tools_data = {
            **self.tools_data,
            tool.tool_id: {
                'tooltip': tool.tool_tip,
                'img': read_icon(path, 'svg+xml'),
                'menu_ind': menu_ind,
                'has_suboptions': has_suboptions,
                'primary': primary,
                'visible': visible
            }
        }

    def vue_select_primary(self, args):
        """
        Activate the primary tool from a given menu index
        """
        menu_ind, tool_id = args
        self._select_tool(tool_id, menu_ind)



from astropy.io import fits
import pytest
import numpy as np
from specutils import SpectralRegion


def test_data_menu_toggles(specviz_helper, spectrum1d):
    # NOTE: this test is adopted from core.tests.test_data_menu.test_data_menu_toggles
    # which should be removed once the old data menu is removed from jdaviz

    # load 2 data entries
    specviz_helper.load_data(spectrum1d, data_label="test")
    new_spec = specviz_helper.get_spectra(apply_slider_redshift=True)["test"]*0.9
    specviz_helper.load_data(new_spec, data_label="test2")

    # check that both are enabled in the data menu
    sv = specviz_helper.viewers['spectrum-viewer']
    dm = sv._obj.data_menu
    assert len(dm._obj.layer_items) == 2
    assert len(dm._obj.visible_layers) == 2

    # disable (hide layer) for second entry
    dm.set_layer_visibility('test2', False)
    assert len(dm._obj.layer_items) == 2
    assert len(dm._obj.visible_layers) == 1

    # add a subset and make sure it appears for the first data entry but not the second
    specviz_helper.plugins['Subset Tools'].import_region(
        SpectralRegion(6000 * spectrum1d.spectral_axis.unit, 6500 * spectrum1d.spectral_axis.unit))

    assert len(dm._obj.layer_items) == 3
    assert len(dm._obj.visible_layers) == 2
    assert len(sv._obj.layers) == 4
    assert sv._obj.layers[2].visible is True   # subset corresponding to first (visible) data entry
    assert sv._obj.layers[3].visible is False  # subset corresponding to second (hidden) data entry

    # enable data layer from menu and subset should also become visible
    dm.toggle_layer_visibility('test2')
    assert np.all([layer.visible for layer in sv._obj.layers])


def test_data_menu_selection(specviz_helper, spectrum1d):
    # load 2 data entries
    specviz_helper.load_data(spectrum1d, data_label="test")
    new_spec = specviz_helper.get_spectra(apply_slider_redshift=True)["test"]*0.9
    specviz_helper.load_data(new_spec, data_label="test2")

    sv = specviz_helper.viewers['spectrum-viewer']
    dm = sv._obj.data_menu

    # no selection by default
    assert len(dm._obj.dm_layer_selected) == 0
    assert len(dm.layer.selected) == 0

    # test layer -> UI sync
    dm.layer.selected = dm.layer.choices[0]
    assert len(dm._obj.dm_layer_selected) == len(dm.layer.selected)

    # test UI -> layer sync
    dm._obj.dm_layer_selected = [0, 1]
    assert len(dm._obj.dm_layer_selected) == len(dm.layer.selected)

    # test that sync remains during layer deletion
    dm._obj.dm_layer_selected = [1]
    assert dm.layer.selected == ['test']
    specviz_helper.app.remove_data_from_viewer("spectrum-viewer", "test2")
    specviz_helper.app.data_item_remove("test2")
    assert len(dm._obj.layer_items) == 1
    assert dm._obj.dm_layer_selected == [0]
    assert dm.layer.selected == ['test']


def test_data_menu_add_remove_data(imviz_helper):
    for i in range(3):
        imviz_helper.load_data(np.zeros((2, 2)) + i, data_label=f'image_{i}', show_in_viewer=False)

    dm = imviz_helper.viewers['imviz-0']._obj.data_menu
    assert len(dm._obj.layer_items) == 0
    assert len(dm.layer.choices) == 0
    assert len(dm._obj.dataset.choices) == 3

    dm.add_data('image_0')
    assert dm.layer.choices == ['image_0']
    assert len(dm._obj.dataset.choices) == 2

    with pytest.raises(ValueError,
                       match="Data labels \\['dne1', 'dne2'\\] not able to be loaded into 'imviz-0'.  Must be one of: \\['image_1', 'image_2'\\]"):  # noqa
        dm.add_data('dne1', 'dne2')

    dm.add_data('image_1', 'image_2')
    assert len(dm.layer.choices) == 3
    assert len(dm._obj.dataset.choices) == 0

    dm.layer.selected = ['image_0']
    dm.remove_from_viewer()
    assert len(dm.layer.choices) == 2
    assert len(dm._obj.dataset.choices) == 1

    dm.layer.selected = ['image_1']
    dm.remove_from_app()
    assert len(dm.layer.choices) == 1
    assert len(dm._obj.dataset.choices) == 1


def test_data_menu_create_subset(imviz_helper):
    imviz_helper.load_data(np.zeros((2, 2)), data_label='image', show_in_viewer=True)

    dm = imviz_helper.viewers['imviz-0']._obj.data_menu
    assert imviz_helper.app.session.edit_subset_mode.edit_subset == []

    dm.create_subset('circle')
    assert imviz_helper.app.session.edit_subset_mode.edit_subset == []
    assert imviz_helper.viewers['imviz-0']._obj.toolbar.active_tool_id == 'bqplot:truecircle'


def test_data_menu_remove_subset(specviz_helper, spectrum1d):
    # load 2 data entries
    specviz_helper.load_data(spectrum1d, data_label="test")
    new_spec = specviz_helper.get_spectra(apply_slider_redshift=True)["test"]*0.9
    specviz_helper.load_data(new_spec, data_label="test2")

    dm = specviz_helper.viewers['spectrum-viewer']._obj.data_menu
    sp = specviz_helper.plugins['Subset Tools']

    sp.import_region(SpectralRegion(6000 * spectrum1d.spectral_axis.unit,
                                    6100 * spectrum1d.spectral_axis.unit),
                     combination_mode='new')
    sp.import_region(SpectralRegion(6000 * spectrum1d.spectral_axis.unit,
                                    6100 * spectrum1d.spectral_axis.unit),
                     combination_mode='new')

    assert dm.layer.choices == ['Subset 2', 'Subset 1', 'test2', 'test']
    dm.layer.selected = ['Subset 1']
    dm.remove_from_viewer()

    # subset visibility is set to false, but still appears in menu (unlike removing data)
    assert dm.layer.choices == ['Subset 2', 'Subset 1', 'test2', 'test']
    assert dm._obj.layer_items[1]['label'] == 'Subset 1'
    # TODO: sometimes appearing as mixed right now, known bug
    assert dm._obj.layer_items[1]['visible'] is not True

    # selection should not have changed by removing subset from viewer
    assert dm.layer.selected == ['Subset 1']
    dm.remove_from_app()
    # TODO: not quite sure why this isn't passing, seems to
    # work on local tests, so may just be async?
    # assert dm.layer.choices == ['test', 'test2', 'Subset 2']


def test_data_menu_dq_layers(imviz_helper):
    hdu_data = fits.ImageHDU(np.zeros(shape=(2, 2)))
    hdu_data.name = 'SCI'
    hdu_dq = fits.ImageHDU(np.zeros(shape=(2, 2), dtype=np.int32))
    hdu_dq.name = 'DQ'
    data = fits.HDUList([fits.PrimaryHDU(), hdu_data, hdu_dq])

    imviz_helper.load_data(data, data_label="image", ext=('SCI', 'DQ'), show_in_viewer=True)

    dm = imviz_helper.viewers['imviz-0']._obj.data_menu
    assert dm.layer.choices == ['image[DQ,1]', 'image[SCI,1]']
    assert len(dm._obj.visible_layers) == 2

    # turning off image (parent) data-layer should also turn off DQ
    dm.set_layer_visibility('image[SCI,1]', False)
    assert len(dm._obj.visible_layers) == 0

    # turning on image (parent) should leave DQ off
    dm.set_layer_visibility('image[SCI,1]', True)
    assert len(dm._obj.visible_layers) == 1

    # turning on DQ (child, when parent off) should show parent
    dm.set_layer_visibility('image[SCI,1]', False)
    dm.set_layer_visibility('image[DQ,1]', True)
    assert len(dm._obj.visible_layers) == 2


@pytest.mark.xfail(reason="known issue")
def test_data_menu_subset_appearance(specviz_helper, spectrum1d):
    # NOTE: this test is similar to above - the subset is appearing in time IF there
    # are two data entries, but not in this case with just one
    specviz_helper.load_data(spectrum1d, data_label="test")

    dm = specviz_helper.viewers['spectrum-viewer']._obj.data_menu
    sp = specviz_helper.plugins['Subset Tools']

    sp.import_region(SpectralRegion(6000 * spectrum1d.spectral_axis.unit,
                                    6100 * spectrum1d.spectral_axis.unit))

    # AssertionError: assert ['Subset 1', 'test'] == ['test', 'Subset 1']
    assert dm.layer.choices == ['test', 'Subset 1']


def test_data_menu_view_info(specviz_helper, spectrum1d):
    # load 2 data entries
    specviz_helper.load_data(spectrum1d, data_label="test")
    new_spec = specviz_helper.get_spectra(apply_slider_redshift=True)["test"]*0.9
    specviz_helper.load_data(new_spec, data_label="test2")

    dm = specviz_helper.viewers['spectrum-viewer']._obj.data_menu
    mp = specviz_helper.plugins['Metadata']
    sp = specviz_helper.plugins['Subset Tools']

    sp.import_region(SpectralRegion(6000 * spectrum1d.spectral_axis.unit,
                                    6100 * spectrum1d.spectral_axis.unit),
                     combination_mode='new')
    sp.import_region(SpectralRegion(6200 * spectrum1d.spectral_axis.unit,
                                    6300 * spectrum1d.spectral_axis.unit),
                     combination_mode='new')

    assert dm.layer.choices == ['Subset 2', 'Subset 1', 'test2', 'test']

    dm.layer.selected = ["test2"]
    dm.view_info()
    assert mp.dataset.selected == "test2"

    dm.layer.selected = ["Subset 2"]
    dm.view_info()
    assert sp.subset.selected == "Subset 2"

    dm.layer.selected = ["test", "test2"]
    with pytest.raises(ValueError, match="Only one layer can be selected to view info"):
        dm.view_info()
